/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher;

import java.io.Writer;
import java.io.IOException;

import java.util.List;
import java.util.LinkedList;
import java.util.Collection;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.RDFUtil;

import dfki.util.config.Config;
import dfki.util.rdfs.RDFS;
import dfki.util.rdf.RDF;

public class AdvancedEdgeExtractor extends AbstractEdgeExtractor
{
    AdvancedEdgeExtractor( ModelProcessor p, GraphBuilder b )
    {
        mModelProcessor = p;
        mGraphBuilder = b;
        mRdfsClasses = new HashMap();
        mRdfsProperties = new HashMap();
    }

    public void extractAllEdges( Config config )
        throws ModelException
    {
        mGraphBuilder.start( config );

        String props = null;
        String hideLiterals = null;
        try
        {
            mRDFSURIs = RDFS.getURIs( config.getPropertyValue( RDFSchemaGrapher.Property.RDFS_NAMESPACE ) );
            mRDFURIs = RDF.syntax( config.getPropertyValue( RDFSchemaGrapher.Property.RDF_NAMESPACE ) );
            mShowRDFSObjects = config.isEnabled( RDFSchemaGrapher.Property.SHOW_RDFS_NODES );
            mProcessingInfo = createNewProcessingInformation();
            props = config.getPropertyValue( RDFSchemaGrapher.Property.SHOW_NON_LITERAL_PROPERTIES );
            hideLiterals = config.getPropertyValue( RDFSchemaGrapher.Property.HIDE_LITERAL_PROPERTIES );
            mShowLiteralPropertiesAsSlot = !config.isEnabled( RDFSchemaGrapher.Property.HIDE_LITERAL_PROPERTIES );
        }
        catch( Config.MultiValueException e )
        {
            RDFSchemaGrapher.debug().error( e );
        }
        catch( RDF.Exception e )
        {
            RDFSchemaGrapher.debug().error( e );
        }

        boolean extractEdges = false;
        mShowNonLieralPropertiesAsSlot = false;

        if (props == null)
            props = RDFSchemaGrapher.SHOW_PROPERTIES_NO;

        if (props.equals( RDFSchemaGrapher.SHOW_PROPERTIES_AS_EDGES ))
        {
            extractEdges = true;
            mShowNonLieralPropertiesAsSlot = false;
        }
        else if (props.equals( RDFSchemaGrapher.SHOW_PROPERTIES_AS_NODE_CONTENT ))
        {
            extractEdges = false;
            mShowNonLieralPropertiesAsSlot = true;
        }
        else if (props.equals( RDFSchemaGrapher.SHOW_PROPERTIES_AS_NODE_CONTENT_AND_EDGES ))
        {
            extractEdges = true;
            mShowNonLieralPropertiesAsSlot = true;
        }
        else /*no*/
        {
            extractEdges = false;
            mShowNonLieralPropertiesAsSlot = false;
        }

        collectObjects();
        extractSubClassOfEdges();
        extractProperties();

        if (extractEdges)
           extractSlotEdges();

        collectNamespaceMappings();
    }


    private void collectObjects()
        throws ModelException
    {
        Collection statements = mModelProcessor.allStatements();
        Iterator i = statements.iterator();
        while( i.hasNext() )
        {
            Statement stmt = (Statement)i.next();

            EdgeExtractor.StatementInfo info = createStatementInfoFor( stmt );
            info.hasBeenConsidered( true );
            mProcessingInfo.addStatementInfo( info );

            String predicateURI = stmt.predicate().getURI();
            String subjectURI = stmt.subject().getURI();
            String objectLabel = stmt.object().getLabel();
            //SubClassOf
            if (predicateURI.equals( rdfsURIs().subClassOf() ))
            {
                RDFSClass.Data d = getRdfsClassData( subjectURI );
                d.superClasses.add( objectLabel );
                setRdfsClassData( subjectURI, d );
            }
            //Domain
            else if (predicateURI.equals( rdfsURIs().domain() ))
            {
                RDFSClass.Data d = getRdfsClassData( objectLabel );
                d.properties.add( subjectURI );
                setRdfsClassData( objectLabel, d );
            }
            //Range
            else if (predicateURI.equals( rdfsURIs().range() ))
            {
                RDFSProperty.Data data = getRdfsPropertyData( subjectURI );
                if (objectLabel.equals( rdfsURIs().literal() ))
                   continue;
                data.ranges.add( objectLabel );
                setRdfsPropertyData( subjectURI, data );
            }
            //Type:class
            //Type:property
            else if (predicateURI.equals( rdfURIs().type().getURI() ))
            {
                if (objectLabel.equals( rdfsURIs()._class() ))
                {
                    RDFSClass.Data d = getRdfsClassData( objectLabel );
                    mGraphBuilder.setNode( mModelProcessor.getAbbreviatedURI( subjectURI ) );
                }
            }
        }
     }

    private void extractSubClassOfEdges()
    {
        Iterator i = rdfsClasses().keySet().iterator();
        while( i.hasNext() )
        {
            String className = (String)i.next();
            RDFSClass.Data data = getRdfsClassData( className );
            Iterator si = data.superClasses.iterator();
            while( si.hasNext() )
            {
                try
                {
                    String sClassName = (String)si.next();
                    if (mModelProcessor.getNamespace( className ).equals( rdfURIs().namespace() ))
                       continue;
                    if (!mShowRDFSObjects
                         && mModelProcessor.getNamespace( sClassName ).equals( rdfsURIs().namespace() ))
                    {
                        continue;
                    }
                    mGraphBuilder.setEdge( mModelProcessor.getAbbreviatedURI( sClassName ),
                                           mModelProcessor.getAbbreviatedURI( className ) );
                }
                catch( ModelException e )
                {
                }
            }
        }
    }

    //properties with class domain
    private void extractSlotEdges()
        throws ModelException
    {
        Iterator i = rdfsClasses().keySet().iterator();
        while( i.hasNext() )
        {
            String className = (String)i.next();
            RDFSClass.Data data = getRdfsClassData( className );
            Iterator pi = data.properties.iterator();
            while( pi.hasNext() )
            {
                String propertyName = (String)pi.next();
                RDFSProperty.Data pData = getRdfsPropertyData( propertyName );
                Iterator ri = pData.ranges.iterator();
                while( ri.hasNext() )
                {
                    String rClassName = (String)ri.next();
                    mGraphBuilder.setEdge( mModelProcessor.getAbbreviatedURI( className ),
                                           mModelProcessor.getAbbreviatedURI( rClassName ),
                                           mModelProcessor.getAbbreviatedURI( propertyName ) );
                }
            }
        }
    }

    private void extractProperties()
        throws ModelException
    {
        Iterator i = rdfsClasses().keySet().iterator();
        while( i.hasNext() )
        {
            String className = (String)i.next();
            RDFSClass.Data data = getRdfsClassData( className );
            Iterator pi = data.properties.iterator();
            while( pi.hasNext() )
            {
                String propertyName = (String)pi.next();
                RDFSProperty.Data pData = getRdfsPropertyData( propertyName );

                Object[] toAdd = null;
                if (mShowLiteralPropertiesAsSlot && pData.ranges.isEmpty())
                {
                    toAdd = new Object[] { mModelProcessor.getAbbreviatedURI( propertyName ) +
                                           ": Literal" };
                }
                else if (mShowNonLieralPropertiesAsSlot && !pData.ranges.isEmpty())
                {
                    String range = (String)pData.ranges.get( 0 );
                    toAdd = new Object[] { mModelProcessor.getAbbreviatedURI( propertyName ) + ": "
                                           + mModelProcessor.getAbbreviatedURI( range ) };
                    mGraphBuilder.setNode( mModelProcessor.getAbbreviatedURI( className ) );
                    mGraphBuilder.setNode( mModelProcessor.getAbbreviatedURI( range ) );
                }
                mGraphBuilder.setEnrichNode( mModelProcessor.getAbbreviatedURI( className ), toAdd );
            }
        }
    }


    private void collectNamespaceMappings()
    {
        try
        {
            Map abbrevs = mModelProcessor.getNamespaceAbbreviations();
            Iterator i = abbrevs.entrySet().iterator();
            while( i.hasNext() )
            {
                Map.Entry entry = (Map.Entry)i.next();
                mGraphBuilder.addNamespaceAbbreviation( (String)entry.getKey(), (String)entry.getValue() );
            }
        }
        catch( ModelException e )
        {
        }
    }

    private Map rdfsClasses()
    {
        return( mRdfsClasses );
    }

    private RDFSClass.Data getRdfsClassData( String name )
    {
        Map clses = rdfsClasses();
        RDFSClass.Data data = (RDFSClass.Data)clses.get( name );
        if (data == null)
        {
            data = new RDFSClass.Data();
            setRdfsClassData( name, data );
        }
        return( data );
    }

    private void setRdfsClassData( String name, RDFSClass.Data data )
    {
        rdfsClasses().put( name, data );
    }


    private RDFSProperty.Data getRdfsPropertyData( String name )
    {
        Map props = rdfsProperties();
        RDFSProperty.Data data = (RDFSProperty.Data)props.get( name );
        if (data == null)
        {
            data = new RDFSProperty.Data();
            setRdfsPropertyData( name, data );
        }
        return( data );
    }

    private void setRdfsPropertyData( String name, RDFSProperty.Data data )
    {
        rdfsProperties().put( name, data );
    }


    private Map rdfsProperties()
    {
        return( mRdfsProperties );
    }

    public void printResultTo( Writer w ) throws IOException
    {
        mGraphBuilder.printGraphTo( w );
    }

    public EdgeExtractor.ProcessingInfo getProcessingInformation()
    {
        return( mProcessingInfo );
    }

    private RDFS.URIs rdfsURIs()
    {
        return( mRDFSURIs );
    }

    private RDF.Syntax rdfURIs()
    {
        return( mRDFURIs );
    }

    private ModelProcessor mModelProcessor;
    private GraphBuilder mGraphBuilder;
    private Map/*String,RDFSClass.Data*/ mRdfsClasses;
    private Map/*String,RDFSProperty.Data*/ mRdfsProperties;
    private EdgeExtractor.ProcessingInfo mProcessingInfo;
    private RDFS.URIs mRDFSURIs;
    private RDF.Syntax mRDFURIs;
    private boolean mShowRDFSObjects;
    private boolean mShowLiteralPropertiesAsSlot;
    private boolean mShowNonLieralPropertiesAsSlot;
}


class RDFSClass
{
/*
    public String getName()
    {
    }
*/
    public static class Data
    {
        Data()
        {
            properties = new LinkedList();
            superClasses = new LinkedList();
        }

        List properties;
        List superClasses;
    }
}


class RDFSProperty
{
/*
    public String getName()
    {
    }
*/
    public static class Data
    {
        Data()
        {
            ranges = new LinkedList();
        }

        List/*String*/ ranges;
        //List domains;
    }
}

class RDFSLiteral
{
    public String getValue()
    {
        return( mValue );
    }
    private String mValue;
}

