/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher;

import java.io.PrintStream;
import java.util.StringTokenizer;

import dfki.util.exec.Executor;

import dfki.util.config.Config;
import dfki.util.Strings;

public class DotExecutor
{
    public final static String DOT_BINARY_PATTERN = "%DOT%";
    public final static String SOURCE_FILE_PATTERN = "%DOT_FILE%";
    public final static String PICTURE_FILE_PATTERN = "%PICTURE_FILE%";
    public final static String PICTURE_TYPE_PATTERN = "%PICTURE_TYPE%";

    public static class Property extends Config.Property
    {
        public final static String MODULE_NAME = "Dot";

        /** Type of output, e.g. gif */
        public final static Property PICTURE_TYPE
                            = new Property( "PictureType", "type of picture to be generated" );

        /** Executing shell: e.g. csh, command.com, cmd.exe */
        public final static Property SHELL
                            = new Property( "Shell", "executing shell"  );

        /** Command string to run the 'dot' tool */
        //public final static Property CALL_PATTERN
        //                    = new Property( "CallPattern", "layouter exec string"  );

        /** Dot executable */
        public final static Property BINARY_NAME
                            = new Property( "BinaryName", "name of the dot executable"  );

        /** Path to font files used by dot */
        public final static Property FONT_PATH
                            = new Property( "FontPath", "path to fonts used by 'dot'"  );

        private final static String PREFIX = MODULE_NAME + Config.Property.SEPARATOR;
        public Property( String id, String descr ) { super( PREFIX + id, descr ); }
    }

    public DotExecutor( String call,
                        String sourceFileName, String outFileName,
                        Config config )
    {
        String type =  null;
        String fontPath = null;
        String binary = null;
        String shell = null;
        try
        {
            shell = config.getPropertyValue( Property.SHELL );
            //call = config.getPropertyValue( Property.CALL_PATTERN );
            binary = config.getPropertyValue( Property.BINARY_NAME );
            type = config.getPropertyValue( Property.PICTURE_TYPE );
            fontPath = config.getPropertyValue( Property.FONT_PATH );
        }
        catch( Config.MultiValueException e )
        {
            RDFSchemaGrapher.debug().error( e );
        }

        if (binary == null)
            binary = "dot";

        call = Strings.replace( call, DOT_BINARY_PATTERN, binary );
        call = Strings.replace( call, SOURCE_FILE_PATTERN, sourceFileName );
        call = Strings.replace( call, PICTURE_FILE_PATTERN, outFileName );
        call = Strings.replace( call, PICTURE_TYPE_PATTERN, type );

        if (fontPath != null)
        {
            mEnvironment = new String[] { "DOTFONTPATH=" + fontPath };
        }

        //prepare shell exectution
        if (shell == null)
           mCommand = new String[] { call };

        StringTokenizer tokenizer = new StringTokenizer( shell, " " );
        mCommand = new String[tokenizer.countTokens()+1];
        for( int i=0; tokenizer.hasMoreElements(); ++i )
        {
            mCommand[i] = tokenizer.nextToken();
        }
        mCommand[mCommand.length-1] = call;
    }

    public boolean execute( PrintStream ps )
    {
        System.err.println( "executing: " + Strings.from( mCommand ) );
        Executor exec = new Executor( mCommand, mEnvironment );
        return( exec.execute( ps ) );
    }

    private Config mConfig;
    private String[] mCommand;
    private String[] mEnvironment;
}