/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher;

import java.io.Writer;
import java.io.IOException;

import java.util.Map;
import java.util.TreeMap;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.Comparator;

import dfki.util.config.Config;
import dfki.util.Strings;

public class DotGraphBuilder implements GraphBuilder
{
    public final static float MIN_RATIO = 0.1f;
    public final static float MAX_RATIO = 10f;

    private static class Edge
    {
        public Edge( String s, String t )
        {
            this( s, t, null );
        }
        public Edge( String s, String t, String l )
        {
            mSource = s;
            mTarget = t;
            mLabel = l;
        }

        public String getSource() { return( mSource ); }
        public String getTarget() { return( mTarget ); }
        public String getLabel() { return( mLabel ); }

        private String mSource;
        private String mTarget;
        private String mLabel;
    }

    private static class Node
    {
        public Node( String l )
        {
            mLabel = l;
            mSubLabels = new LinkedList();
        }
        public String getLabel() { return( mLabel ); }
        public Collection getSubLabels() { return( mSubLabels ); }
        public void addSubLabel( Object[] l ) { mSubLabels.add( l ); }
        public boolean equals( Object other ) { return( mLabel.equals( ((Node)other).mLabel )); }

        private Collection/*Object[]*/ mSubLabels;
        private String mLabel;
    }


    public DotGraphBuilder( String graphName )
    {
        mGraphName = graphName;
    }

    public void start( Config config )
    {
        reset();

        String props = null;
        String ratioEnabled = null;
        String ratio = null;
        boolean hideLiterals = true;
        try
        {
            mNodeColor = config.getPropertyValue( RDFSchemaGrapher.Property.NODE_COLOR );
            mNodeBackgroundColor = config.getPropertyValue( RDFSchemaGrapher.Property.NODE_BACKGROUND_COLOR );
            mIsAEdgeColor = config.getPropertyValue( RDFSchemaGrapher.Property.ISA_EDGE_COLOR );
            mPropertyEdgeColor = config.getPropertyValue( RDFSchemaGrapher.Property.PROPERTY_EDGE_COLOR );
            mPrintNamespaces = config.isEnabled( RDFSchemaGrapher.Property.PRINT_NAMESPACES );
            props = config.getPropertyValue( RDFSchemaGrapher.Property.SHOW_NON_LITERAL_PROPERTIES );
            ratioEnabled = config.getPropertyValue( RDFSchemaGrapher.Property.ENABLE_RATIO );
            ratio = config.getPropertyValue( RDFSchemaGrapher.Property.ASPECT_RATIO );
            hideLiterals = config.isEnabled( RDFSchemaGrapher.Property.HIDE_LITERAL_PROPERTIES );
            mUseUMLStyle = config.isEnabled( RDFSchemaGrapher.Property.UML_STYLE_NODES );
        }
        catch( Config.MultiValueException e )
        {
            RDFSchemaGrapher.debug().error( e );
        }


        if (props == null)
            props = RDFSchemaGrapher.SHOW_PROPERTIES_NO;

        mNodeShape = "ellipse";

        if (!hideLiterals
            || props.equalsIgnoreCase( RDFSchemaGrapher.SHOW_PROPERTIES_AS_NODE_CONTENT )
            || props.equalsIgnoreCase( RDFSchemaGrapher.SHOW_PROPERTIES_AS_NODE_CONTENT_AND_EDGES )
            || mUseUMLStyle )
        {
            mNodeShape = "record";
        }


        //ratio check
        try
        {
            if (ratioEnabled != null)
            {
                float f = Float.parseFloat( ratio );
                if (f>MAX_RATIO || f<MIN_RATIO)
                {
                    ratio = null;
                }
            }
            else
            {
                ratio = null;
            }
        }
        catch( Throwable ex )
        {
            ratio = null;
        }
        mRatio = ratio;

        if (mNodeColor == null) mNodeColor = RDFSchemaGrapher.DEFAULT_NODE_COLOR;

        //print out some config info
        RDFSchemaGrapher.debug().message( RDFSchemaGrapher.Property.ENABLE_RATIO.getId()
                                            + " = " + (ratioEnabled==null?"no":ratioEnabled) );
        if (mRatio != null)
        {
            RDFSchemaGrapher.debug().message( RDFSchemaGrapher.Property.ASPECT_RATIO.getId()
                                                + " = " + mRatio );
        }
    }

    private void reset()
    {
        mEdgeList = new LinkedList();
        mNodes = new TreeMap();
        mNamespaceAbbrevs = new TreeMap();
    }


    public void setNode( String name )
    {
        //just create a node if not present yet
        Node n = getNode( name );
    }

    public void setEdge( String source, String target )
    {
        mEdgeList.add( new Edge( source, target ) );
    }

    public void setEdge( String source, String target, String label )
    {
        mEdgeList.add( new Edge( source, target, label ) );
    }

    public void setEnrichNode( String node, Object[] toAdd )
    {
        Node n = getNode( node );
        if (toAdd == null)
           return;
        n.addSubLabel( toAdd );
    }

    public void addNamespaceAbbreviation( String namespace, String abbrev )
    {
        mNamespaceAbbrevs.put( abbrev, namespace );
    }

    public void printGraphTo( Writer w ) throws IOException
    {
        startWrite( w );

        //Legende
        if (!mNamespaceAbbrevs.isEmpty() && mPrintNamespaces)
        {
            w.write( "\n\n  \"Namespaces:\\n" );
            Iterator i = mNamespaceAbbrevs.keySet().iterator();
            while( i.hasNext() )
            {
                String abbr = (String)i.next();
                w.write( abbr + " = " + mNamespaceAbbrevs.get( abbr ) + "\\l" );
            }
            w.write( "\" [shape=record,style=filled,color=lightgrey,fontsize=9,fontcolor=black]\n" );
        }

        //edges
        boolean generatedSomeNodes = false;
        Iterator i = mEdgeList.iterator();
        while( i.hasNext() )
        {
            Edge edge = (Edge)i.next();
            String label = edge.getLabel();
            //if (label != null && !mPrintPropertyEdges)
            //   continue;
            w.write( "\n  \"" + escapeNodeLabel( edge.getSource() )
                      + "\" -> \"" + escapeNodeLabel( edge.getTarget() ) + "\"" );
            if (label != null)
            {
                w.write( " [fontcolor=" + mPropertyEdgeColor + ",label=\"" + label
                               + "\",style=dashed,color=" + mPropertyEdgeColor + ",dir=forward]" );
            }
            generatedSomeNodes = true;
        }
        w.write( "\n" );

        //enriched nodes (with property entrys)
        i = mNodes.values().iterator();
        while( i.hasNext() )
        {
            Node node = (Node)i.next();
            Collection subLabels = node.getSubLabels();
            subLabels = sortSubLabels( subLabels );

            generatedSomeNodes = true;
            String props = "";
            Iterator li = subLabels.iterator();
            while( li.hasNext() )
            {
                Object[] subLabel = (Object[])li.next();

                //props += "{";
                for( int sli=0; sli<subLabel.length; ++sli )
                {
                    if (sli != 0)
                       props += "|";
                    if (mUseUMLStyle)
                       props += "+ ";
                    props += escapeNodeLabel( subLabel[sli].toString() );
                }
                /* slots seperated and centered
                props += "}";
                if (li.hasNext())
                   props += "|";
                */

                /* slots in one box and left aligned*/
                if (mUseUMLStyle)
                    props += "\\l";
                else
                    props += "\\n";
            }

            String label = escapeNodeLabel( node.getLabel() );
            w.write( "\n \"" + label + "\"" );
            if (props.length() == 0)
                continue;
            w.write( "[label=\"{" + label + "\\n|" + props + "}\"]" );
        }


        if (!generatedSomeNodes)
        {
            w.write( "\n\"Sorry, no nodes generated\" [shape=plaintext,fontsize=18,fontcolor=black]" );
            RDFSchemaGrapher.debug().message( "No nodes generated? Check, if you are using the right rdfs namespace" );
            RDFSchemaGrapher.debug().message( "... and remember: this tool visualizes rdfs NOT rdf" );
        }

        endWrite( w );
        w.flush();
    }

    private void startWrite( Writer w ) throws IOException
    {
        w.write( "digraph " + mGraphName + "\n{\n" );
        w.write( "  node [fontname=Arial," );
        //w.write( "  node [" );
        if (!mNodeBackgroundColor.equals( RDFSchemaGrapher.NODE_BACKGROUND_COLOR_NONE ))
        {
            w.write( "style=filled,color=" + mNodeBackgroundColor + "," );
        }
        w.write( "fontcolor=" + mNodeColor + ",shape=" + mNodeShape + "]\n" );
        w.write( "  edge [fontname=Arial, dir=back"
        //w.write( "  edge [dir=back"
                      + ",color=" + mIsAEdgeColor
                      + ",fontcolor=" + mIsAEdgeColor + "]\n" );
        if (mRatio != null)
        {
            w.write( "   ratio=\"" + mRatio + "\"" + "\n" );
        }
    }

    private void endWrite( Writer w ) throws IOException
    {
        w.write( "\n}\n" );
    }

    private Node getNode( String label )
    {
        Node n = (Node)mNodes.get( label );
        if (n == null)
        {
            n = new Node( label );
            mNodes.put( label, n );
        }
        return( n );
    }

    private Collection sortSubLabels( Collection c )
    {
        List l = new ArrayList( c );
        Collections.sort( l,
              new Comparator()
              {
                public int compare( Object o1, Object o2 )
                {
                    Object elt1 = ((Object[])o1)[0];
                    Object elt2 = ((Object[])o2)[0];
                    return( (Comparable)elt1).compareTo( elt2 );
                }
                public boolean equals( Object o ) { return( this == o ); }
              } );
        return( l );
    }

    private String escapeNodeLabel( String label )
    {
        label = replaceAll( label, "\\", "\\\\" );
        label = replaceAll( label, "{", "\\{" );
        label = replaceAll( label, "}", "\\}" );
        label = replaceAll( label, "|", "\\|" );
        label = replaceAll( label, "\"", "\\\"" );
        label = replaceAll( label, "\n", " " );

        return( label );
    }

    public String replaceAll( String str, final String pattern, final String replace )
    {
        int pos = 0;
        while( pos!=-1 )
        {
            pos = str.indexOf( pattern, pos );
            if (pos != -1)
            {
                str = Strings.replace( str, pattern, replace, pos );
                pos += replace.length();
            }
        }
        return( str );
    }

    private String mGraphName;
    private LinkedList mEdgeList;
    private Map mNodes;
    private Map mNamespaceAbbrevs;

    private String mNodeColor;
    private String mNodeBackgroundColor;
    private String mPropertyEdgeColor;
    private String mIsAEdgeColor;
    private String mNodeShape;
    private String mRatio;
    private boolean mUseUMLStyle = false;

    private boolean mPrintNamespaces;
    //private boolean mPrintPropertyEdges;
    //private boolean mPrintPropertiesInNodes;
}