/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Copyright:    Copyright (c) DFKI GmbH<p>
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher;

import java.util.Collection;
import java.util.Hashtable;
import java.util.Map;
import java.util.Iterator;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.RDFUtil;

import dfki.util.config.Config;
import dfki.util.rdf.RDF;
import dfki.util.rdfs.RDFS;

public class ModelProcessor
{
    public ModelProcessor( Model m, Config config )
    {
        if (m == null)
           throw( new NullPointerException() );

        try
        {
            mModel = m;
            mAllStatements = null;

            mEnableNamespaces = config.isEnabled( RDFSchemaGrapher.Property.PRINT_NAMESPACES );

            mRDFSNamespace = RDFS.getURIs( config.getPropertyValue( RDFSchemaGrapher.Property.RDFS_NAMESPACE ) ).namespace();
            mRDFNamespace = RDF.syntax( config.getPropertyValue( RDFSchemaGrapher.Property.RDF_NAMESPACE ) ).namespace();
        }
        catch( Config.MultiValueException e )
        {
            RDFSchemaGrapher.debug().error( e );
        }
        catch( RDF.Exception e )
        {
            RDFSchemaGrapher.debug().error( e );
        }
    }

    /**
     * Liefert alle Statements (Aussagen) des aktuellen Modells.
     */
    public Collection allStatements()
    {
        if (mAllStatements != null)
           return( mAllStatements );

        try
        {
            mAllStatements = RDFUtil.getStatementVector( mModel );
            //System.out.println( mAllStatements );
        }
        catch( ModelException ignore )
        {
        }

        return( mAllStatements );
    }


    /**
     * Extrahiert alle namespaces aus dem aktuellen Modell.
    public Collection extractNamespaces() throws ModelException
    {
        Map namespacesAbbrevs = getNamespaceAbbreviations();
        return( namespacesAbbrevs.keySet() );
    }
     */

    public String getAbbreviatedURI( String uri ) throws ModelException
    {
        String namespace = getNamespace( uri );
        addNamespace( namespace );
        String name = getName( uri );
        if (!mEnableNamespaces)
           return( name );

        String abbr = getAbbreviatedNamespace( namespace );
        if (abbr == null)
           return( name );
        else
           return( abbr + ":" + name );
    }

    private String getAbbreviatedNamespace( String namespace ) throws ModelException
    {
        addNamespace( namespace );
        Map abbrs = getNamespaceAbbreviations();
        return( (String)abbrs.get( namespace ) );
    }

    public String getName( String uri )
    {
        return( RDFUtil.guessName( uri ) );
    }

    public String getNamespace( String uri )
    {
        String namespace = RDFUtil.guessNamespace( uri );
        if (namespace.endsWith( "##" ))
        {
            namespace = namespace.substring( 0, namespace.length()-1 );
        }
        addNamespace( namespace );
        return( namespace );
    }

    public Map getNamespaceAbbreviations() throws ModelException
    {
        if (mNamespaceAbbreviations != null)
           return( mNamespaceAbbreviations );

        mNamespaceAbbreviations = new Hashtable();
        getNamespace( mRDFSNamespace );
        getNamespace( mRDFNamespace );
/*
        mNamespaceCounter = 0;
        Collection statements = allStatements();
        Iterator i = statements.iterator();
        while( i.hasNext() )
        {
            Statement stmt = (Statement)i.next();
            String namespace = null;
            String abbr = null;

            if (RDF.type.equals( stmt.predicate() )) // RDFMS.type
            {
                getNamespace( stmt.object().getLabel() );
            }
            else if (RDFS.subClassOf.equals( stmt.predicate() )
                     || RDFS.subPropertyOf.equals( stmt.predicate() ))
            {
            }

            getNamespace( stmt.predicate().getLabel() );


            getNamespace( stmt.subject().getLabel() );
            if (stmt.object() instanceof Resource )
            {
                getNamespace( stmt.object().getLabel() );
            }
        }
*/
        return( mNamespaceAbbreviations );
    }

    private void addNamespace( String namespace )
    {
        try
        {
            addNamespace( getNamespaceAbbreviations(), namespace );
        }
        catch( Exception ignore )
        {
        }
    }

    private void addNamespace( Map abbreviationMap, String namespace )
    {
        String shortcut = null;

        if (namespace.length() == 0)
           return;

        //schon eingetragen?
        if (abbreviationMap.containsKey( namespace ))
           return;

        //neue Abkuerzung ermitteln
        if (namespace.equals( "#" ))
            shortcut = "local";
        else if (mRDFNamespace.equals( namespace ))
           shortcut = "rdf";
        else if(mRDFSNamespace.equals( namespace ))
           shortcut = "rdfs";
        else
           shortcut = (mNamespaceCounter < (int)'s' - 1)
                             ? String.valueOf((char)((int)'a' + mNamespaceCounter++))
                             : "n" + (++mNamespaceCounter - (int)'s' + 1);

        abbreviationMap.put( namespace, shortcut );
    }

    private Model mModel;
    private Collection mAllStatements;
    private Map mNamespaceAbbreviations;
    private int mNamespaceCounter;
    private boolean mEnableNamespaces;
    private String mRDFSNamespace;
    private String mRDFNamespace;
}