/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher;

import java.io.File;
import java.io.Writer;
import java.io.Reader;
import java.io.FileReader;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;

import java.util.Collection;
import java.util.Collections;


import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import org.xml.sax.SAXException;

import dfki.util.application.OptionReader;
import dfki.util.application.Option;
import dfki.util.application.ReturnCode;

import dfki.util.rdf.RDF;

import dfki.util.config.Config;
import dfki.util.config.ConfigHolder;
import dfki.util.config.SystemPropertyReader;

import dfki.util.debug.Debug;

import org.w3c.rdf.examples.ParseAndSerialize;


public class RDFSchemaGrapher extends ConfigHolder
{
    public final static String MODULE_NAME = "RDFSViz";

    //values for property SHOW_NON_LITERAL_PROPERTIES
    public final static String SHOW_PROPERTIES_NO = "no";
    public final static String SHOW_PROPERTIES_AS_NODE_CONTENT = "asNodeContent";
    public final static String SHOW_PROPERTIES_AS_EDGES = "asEdges";
    public final static String SHOW_PROPERTIES_AS_NODE_CONTENT_AND_EDGES = "asNodeContentAndEdges";

    public final static String NODE_BACKGROUND_COLOR_NONE = "none";

    //color defaults
    public final static String DEFAULT_NODE_COLOR = "black";
    public final static String DEFAULT_NODE_BACKGROUND_COLOR = "none";
    public final static String DEFAULT_NODE_ISA_EDGE_COLOR = "black";
    public final static String DEFAULT_PROPERTY_EDGE_COLOR = "grey70";

    //grapher properties
    public final static class Property extends Config.Property
    {
        public final static Property PRINT_NAMESPACES
                            = new Property( "PrintNamespaces",
                                            "Add namespaces to extracted class and property names" );
        public final static Property SHOW_NON_LITERAL_PROPERTIES
                            = new Property( "ShowNonLiteralProperties",
                                            "How to show properties with class domain" );
        public final static Property HIDE_LITERAL_PROPERTIES
                            = new Property( "HideLiteralProperties",
                                            "Do not display literal-valued properties" );
        public final static Property UML_STYLE_NODES
                            = new Property( "UMLStyleNodes",
                                            "Print the graph in UML syle" );

        //color properties
        public final static Property NODE_COLOR
                            = new Property( "NodeColor",
                                            "Color of generated nodes" );
        public final static Property NODE_BACKGROUND_COLOR
                            = new Property( "NodeBackgroundColor",
                                            "Background color of generated nodes" );
        public final static Property ISA_EDGE_COLOR
                            = new Property( "IsAEdgeColor",
                                            "Color of subclass hierarchy edges" );
        public final static Property PROPERTY_EDGE_COLOR
                            = new Property( "PropertyEdgeColor",
                                            "Color of property edges" );
        //size
        public final static Property ASPECT_RATIO
                            = new Property( "AspectRatio",
                                            "Aspect ratio" );
        public final static Property ENABLE_RATIO
                            = new Property( "EnableRatio",
                                            "Enable the aspect ratio settings" );

        //additional infos
        public final static Property SHOW_RDFS_NODES
                            = new Property( "ShowRDFSNodes",
                                            "Show rdfs nodes" );

        //namespace settings
        public final static Property RDFS_NAMESPACE
                            = new Property( "RDFSNamespace",
                                            "Used rdf schema namespace" );
        public final static Property RDF_NAMESPACE
                            = new Property( "RDFNamespace",
                                            "Used rdf namespace" );

        //verbosity
        public final static Property VERBOSE
                            = new Property( "Verbose",
                                            "Print additional information" );

        public final static Property[] ALL = { PRINT_NAMESPACES, SHOW_NON_LITERAL_PROPERTIES,
                 HIDE_LITERAL_PROPERTIES, NODE_COLOR, NODE_BACKGROUND_COLOR, ISA_EDGE_COLOR,
                 PROPERTY_EDGE_COLOR, ASPECT_RATIO, ENABLE_RATIO, SHOW_RDFS_NODES,
                 RDFS_NAMESPACE, RDF_NAMESPACE };

        private final static String PREFIX = MODULE_NAME + Config.Property.SEPARATOR;

        private Property( String id, String descr ) { super( PREFIX + id, descr ); }
    }

    public static class Exception extends java.lang.Exception
    {
        Exception( String msg ) { super( /*RDFSchemaGrapher.class.getName() + ": " +*/ msg ); }
    }

    public static Debug debug()
    {
        return( Debug.forThread( MODULE_NAME ) );
    }

    public void setConfig( Config c )
    {
        c.setDefaultPropertyValue( Property.NODE_COLOR, DEFAULT_NODE_COLOR );
        c.setDefaultPropertyValue( Property.ISA_EDGE_COLOR, DEFAULT_NODE_ISA_EDGE_COLOR );
        c.setDefaultPropertyValue( Property.NODE_BACKGROUND_COLOR, DEFAULT_NODE_BACKGROUND_COLOR );
        c.setDefaultPropertyValue( Property.PROPERTY_EDGE_COLOR, DEFAULT_PROPERTY_EDGE_COLOR );


        super.setConfig( c );
    }
    //--------------------------------------------

    public RDFSchemaGrapher( Reader reader, Writer writer )
    {
        mWriter = writer;
        mReader = reader;
        mEdgeExtractor = null;
    }

    public ModelProcessor createModelProcessor()
    {
        return( new ModelProcessor( mModel, config() ) );
    }

    public GraphBuilder createGraphBuilder()
    {
        return( new DotGraphBuilder( "RDFSGraph" ) );
    }

    public EdgeExtractor createEdgeExtractor()
    {
        ModelProcessor modelProcessor = createModelProcessor();
        GraphBuilder graphBuilder = createGraphBuilder();
        return( new AdvancedEdgeExtractor( modelProcessor, graphBuilder ) );
    }


    public void readRDFS( String uri )
       throws IOException,
              SAXException,
              ModelException
    {
        RDFFactory f = new RDFFactoryImpl();
        mModel = f.createModel();

        RDF.parse( uri, mReader, f.createParser(), mModel );

        //RDFUtil.printStatements(mModel, System.out);
    }

    public void setLocalURI( String uri )
    {
        if (!uri.endsWith( "#" ))
        {
            uri = uri + "#";
        }
        mLocalURI = uri;
    }


    public void run()
        throws Exception
    {
        try
        {
            if (config().isEnabled( Property.VERBOSE ))
               debug().setDebugMode( Debug.Mode.NORMAL );
            else
               debug().setDebugMode( Debug.Mode.NODEBUG );

            debug().message( config().printProperty( RDFSchemaGrapher.Property.RDFS_NAMESPACE ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.SHOW_NON_LITERAL_PROPERTIES ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.HIDE_LITERAL_PROPERTIES ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.SHOW_RDFS_NODES ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.PRINT_NAMESPACES ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.NODE_COLOR ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.NODE_BACKGROUND_COLOR ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.ISA_EDGE_COLOR ) );
            debug().message( config().printProperty( RDFSchemaGrapher.Property.PROPERTY_EDGE_COLOR ) );

            readRDFS( mLocalURI );

            mEdgeExtractor = createEdgeExtractor();
            mEdgeExtractor.extractAllEdges( config() );

            mEdgeExtractor.printResultTo( mWriter );
        }
        catch( java.lang.Exception e )
        {
            e.printStackTrace( System.err );
            throw( new Exception( e.toString() ) );
        }
    }

    public EdgeExtractor.ProcessingInfo getExtractorProcessingInfo()
    {
        if (mEdgeExtractor == null)
           return( null);

        return(  mEdgeExtractor.getProcessingInformation() );
    }

    private Model mModel;
    private Writer mWriter;
    private Reader mReader;
    private String mLocalURI;
    private EdgeExtractor mEdgeExtractor;

    // ---------------------------------------------------------

    public static class Test
    {
        public static void main( String[] args ) throws java.lang.Exception
        {
            System.err.println( Test.class.getName() + " <rdfFile> [ <outFile> ]" );
            Writer w = new PrintWriter( System.out );
            if (args.length >=2 )
            {
                w = new FileWriter( args[1] );
            }

            Reader r = new FileReader( args[0] );

            System.getProperties().put( RDFSchemaGrapher.Property.PRINT_NAMESPACES.getId(), "on" );
            System.getProperties().put( RDFSchemaGrapher.Property.SHOW_NON_LITERAL_PROPERTIES.getId(), "asEdges" );
            System.getProperties().put( RDFSchemaGrapher.Property.NODE_COLOR.getId(), "green" );
            System.getProperties().put( RDFSchemaGrapher.Property.PROPERTY_EDGE_COLOR.getId(), "red" );
            System.getProperties().put( RDFSchemaGrapher.Property.RDFS_NAMESPACE.getId(), "http://www.w3.org/TR/1999/zxc#" );
            //System.getProperties().put( RDFSchemaGrapher.Property.RDF_NAMESPACE.getId(), "red" );
            System.getProperties().put( RDFSchemaGrapher.Property.SHOW_RDFS_NODES.getId(), "on" );

            RDFSchemaGrapher g = new RDFSchemaGrapher( r, w );
            Config c = Config.cfg();
            c.addSettings( new SystemPropertyReader() );
            g.setConfig( c );
            g.setLocalURI( args[0] );
            g.run();
        }
    }
}