/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher.app;

import java.io.File;
import java.io.Writer;
import java.io.Reader;
import java.io.InputStreamReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;

import java.net.URL;
import java.net.URLConnection;

import dfki.frodo.tools.rdf_schema_grapher.RDFSchemaGrapher;

import dfki.util.application.OptionReader;
import dfki.util.application.Option;
import dfki.util.application.ReturnCode;

import dfki.util.config.Config;
import dfki.util.config.SystemPropertyReader;
import dfki.util.config.XMLConfigReader;

import dfki.util.debug.Debug;

import dfki.util.io.Files;

import dfki.util.xml.XML;

/**
 * Command line version of the gapher tool (currently without visualization).
 * This class provides for generating 'dot' files out of RDF schema files.
 *
 */

public class RDFSchemaGrapherApp
{
    //command line options
    public final static int _RDFS = 1;
    public final static int _CONFIG = 2;
    public final static int _RDFS_NAMESPACE = 3;
    public final static int _NODE_COLOR = 4;
    public final static int _ISA_EDGE_COLOR = 5;
    public final static int _PROPERTY_EDGE_COLOR = 6;
    public final static int _PRINT_NAMESPACES = 7;
    public final static int _SAX_PARSER = 8;
    public final static int _VERBOSE = 9;
    public final static int _OUTPUT = 10;
    public final static int _PROPERTIES = 11;
    public final static int _SHOW_RDFS_NODES = 12;
    public final static int _ASPECT_RATIO = 13;
    public final static int _HIDE_LITERAL_PROPERTIES = 14;

    /** The URL of the source rdfs file */
    public final static Option RDFS = new Option( "-rdfs", _RDFS,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NECESSARY,
                                             null,
                                             "URL of the RDFS source file (local files must start with file:///)" );
    /** Grapher configuration file (not supported yet) */
    public final static Option CONFIG = new Option( "-config", _CONFIG,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "XML Grapher configuration file" );
    /** RDFS file to process */
    public final static Option RDFS_NAMESPACE = new Option( "-RDFSNamespace", _RDFS_NAMESPACE,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             "http://www.w3.org/TR/1999/PR-rdf-schema-19990303#",
                                             "Use this rdfs namespace" );
    /** Node color */
    public final static Option NODE_COLOR = new Option( "-NodeColor", _NODE_COLOR,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "Node color" );
    /** Color of subclass hierarchy edges */
    public final static Option ISA_EDGE_COLOR = new Option( "-IsAEdgeColor", _ISA_EDGE_COLOR,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "Color of IsA edges" );
    /** Color of property edges */
    public final static Option PROPERTY_EDGE_COLOR = new Option( "-PropertyEdgeColor", _PROPERTY_EDGE_COLOR,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "Color of property edges" );
    /** Write generated output to this file */
    public final static Option OUTPUT = new Option( "-o", _OUTPUT,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "Write generated output to this file" );
    /** Generate namespace information */
    public final static Option PRINT_NAMESPACES = new Option( "-PrintNamespaces", _PRINT_NAMESPACES,
                                             Option.WITHOUT_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "Generate namespace information" );
    /** Show property edges */
    public final static Option PROPERTIES = new Option( "-Properties", _PROPERTIES,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             RDFSchemaGrapher.SHOW_PROPERTIES_AS_EDGES,
                                             "display non-literal properties: no, asEdges, asNodeContent, asNodeContentAndEdges" );

    /** Show rdfs nodes */
    public final static Option SHOW_RDFS_NODES = new Option( "-ShowRDFSNodes", _SHOW_RDFS_NODES,
                                             Option.WITHOUT_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             null,
                                             "Show rdfs nodes" );
    /** Use this SAXParser to read the XML */
    public final static Option SAX_PARSER = new Option( "-SAXParser", _SAX_PARSER,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY,
                                             "org.apache.xerces.parsers.SAXParser",
                                             "Use this SAXParser to read the XML" );
    /** Enable message printing (not supported yet) */
    public final static Option VERBOSE = new Option( "-v", _VERBOSE,
                                             Option.WITHOUT_VALUE, Option.UNIQUE, Option.NOT_NECESSARY, null,
                                             "Enable message printing" );

    public final static Option ASPECT_RATIO = new Option( "-AspectRatio", _ASPECT_RATIO,
                                             Option.NEEDS_VALUE, Option.UNIQUE, Option.NOT_NECESSARY, null,
                                             "Sets the aspect ratio of the generated graph" );

    public final static Option HIDE_LITERAL_PROPERTIES = new Option( "-HideLiteralProperties", _HIDE_LITERAL_PROPERTIES,
                                             Option.WITHOUT_VALUE, Option.UNIQUE, Option.NOT_NECESSARY, null,
                                             "hides literal-valued properties" );

    public final static Option[] OPTIONS = { RDFS, CONFIG, RDFS_NAMESPACE,
                                             NODE_COLOR, ISA_EDGE_COLOR, PROPERTY_EDGE_COLOR,
                                             OUTPUT, PRINT_NAMESPACES, SAX_PARSER, PROPERTIES,
                                             VERBOSE, ASPECT_RATIO, HIDE_LITERAL_PROPERTIES };


    public static void main( String[] args )
    {
        OptionReader optionReader = new OptionReader( OPTIONS );
        try
        {
            if (args.length == 0)
            {
                System.err.println( "Usage:");
                System.err.println( RDFSchemaGrapherApp.class.getName() + " { <options> }" );
                System.err.println( optionReader.getOverview() );
                System.exit( 0 );
            }

            //read command line args
            optionReader.readArgs( args );
        }
        catch( OptionReader.Exception e )
        {
            e.printStackTrace( System.out );
            System.exit( ReturnCode.PARAMETER_ERROR );
        }


        try
        {
            Config config = Config.cfg( new SystemPropertyReader() );

            RDFSchemaGrapher.debug().setDebugMode( Debug.Mode.ERROR );
            if (optionReader.hasOption( VERBOSE ))
            {
                RDFSchemaGrapher.debug().setDebugMode( Debug.Mode.FULL );
                config.setPropertyValue( RDFSchemaGrapher.Property.VERBOSE, Config.ENABLED );
            }

            //XML parser
            String parserClassName = optionReader.getOptionValue( SAX_PARSER );
            XML.Setup.setSAXParserClassName( parserClassName );

            String xmlConfigFile = optionReader.getOptionValue( CONFIG );
            if (xmlConfigFile != null)
            {
                config.addSettings( new XMLConfigReader( xmlConfigFile ) );
            }

            //Color settings
            String col = optionReader.getOptionValue( NODE_COLOR );
            config.setPropertyValue( RDFSchemaGrapher.Property.NODE_COLOR, col );

            col = optionReader.getOptionValue( ISA_EDGE_COLOR );
            config.setPropertyValue( RDFSchemaGrapher.Property.ISA_EDGE_COLOR, col );

            col = optionReader.getOptionValue( PROPERTY_EDGE_COLOR );
            config.setPropertyValue( RDFSchemaGrapher.Property.PROPERTY_EDGE_COLOR, col );

            String rdfsNamespace = optionReader.getOptionValue( RDFS_NAMESPACE );
            if (rdfsNamespace != null)
               config.setPropertyValue( RDFSchemaGrapher.Property.RDFS_NAMESPACE, rdfsNamespace );

            if (optionReader.hasOption( PRINT_NAMESPACES ))
               config.setPropertyValue( RDFSchemaGrapher.Property.PRINT_NAMESPACES, Config.ENABLED );

            //show RDFS properties?
            String showProps = optionReader.getOptionValue( PROPERTIES );
            config.setPropertyValue( RDFSchemaGrapher.Property.SHOW_NON_LITERAL_PROPERTIES, showProps );

            //aspect ratio
            String ratio = optionReader.getOptionValue( ASPECT_RATIO );
            config.setPropertyValue( RDFSchemaGrapher.Property.ASPECT_RATIO, ratio );

            //hide literal-valued properties?
            if (optionReader.hasOption( HIDE_LITERAL_PROPERTIES ))
                config.setPropertyValue( RDFSchemaGrapher.Property.HIDE_LITERAL_PROPERTIES, Config.ENABLED );


            //Input file
            URL rdfsURL = new URL( optionReader.getOptionValue( RDFS ) );
            Reader rdfsReader = new InputStreamReader( rdfsURL.openConnection().getInputStream() );

            String outputFilename = optionReader.getOptionValue( OUTPUT );
            outputFilename = Files.getAbsoluteFilename( outputFilename );


            Writer writer = new PrintWriter( System.out );
            if (outputFilename != null)
            {
                writer = new FileWriter( outputFilename );
            }

            RDFSchemaGrapher g = new RDFSchemaGrapher( rdfsReader, writer );
            g.setConfig( config );
            g.setLocalURI( rdfsURL.toString() );
            g.run();
        }
        catch( Exception e )
        {
            RDFSchemaGrapher.debug().error( e );
            System.exit( ReturnCode.INTERNAL_ERROR );
        }
    }
}