/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Copyright:    Copyright (c) DFKI GmbH<p>
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.frodo.tools.rdf_schema_grapher.servlet;

import javax.servlet.ServletConfig;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import java.io.Reader;
import java.io.Writer;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.File;
import java.io.PrintWriter;
import java.io.IOException;

import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.MalformedURLException;
import java.net.UnknownHostException;

import java.util.Iterator;
import java.util.Properties;
import java.util.Date;
import java.util.Locale;

import java.text.DateFormat;

import org.w3c.rdf.model.Statement;


import dfki.util.config.Config;
import dfki.util.config.ConfigReader;
import dfki.util.config.ServletInitParameterReader;
import dfki.util.config.ServletRequestReader;
import dfki.util.config.SystemPropertyReader;

import dfki.util.debug.BufferHandler;

import dfki.frodo.tools.rdf_schema_grapher.RDFSchemaGrapher;
import dfki.frodo.tools.rdf_schema_grapher.QueryCache;
import dfki.frodo.tools.rdf_schema_grapher.DotExecutor;
import dfki.frodo.tools.rdf_schema_grapher.EdgeExtractor;

import dfki.util.servlet.DebugHttpServletRequest;
import dfki.util.servlet.DebugHttpServletResponse;

import dfki.util.http.HTTP;

import dfki.util.html.HTML;

/**
 * @see dfki.frodo.tools.rdf_schema_grapher.RDFSchemaGrapher.Property
 * @see dfki.frodo.tools.rdf_schema_grapher.DotExecutor.Property
 */

public class GrapherServlet extends HttpServlet
{
    public static class Property extends Config.Property
    {
        //servlet specific init settings
        /** dir to store the generated pics */
        public final static Property PICTURE_DIR
                              = new Property( "PictureDir", "directory for the generated pictures" );

        /** dir to store rdfs files */
        public final static Property RDFS_DIR
                              = new Property( "RDFSDir", "Directory for the rdfs file to process" );

        /** dir to store the generated dot files */
        public final static Property DOT_DIR
                              = new Property( "DotDir", "Directory for the generated dot files" );

        /** size of the picture cache */
        public final static Property CACHE_SIZE
                              = new Property( "CacheSize", "Size of the picture cache" );

        /** URL to access the generated pics through a HTTP server */
        public final static Property PICTURE_URL
                              = new Property( "PictureURL", "URL to read the generated pictures" );

        /** URL to access the generated dot files through a HTTP server */
        public final static Property DOT_URL
                              = new Property( "DotURL", "URL to read the generated dot files" );

        //public final static String RAW_RESULT_HTML_PAGE = "ResultHTMLPage";
        /** dir to store additional info of the generation tasks */
        public final static Property INFO_DIR
                              = new Property( "InfoDir", "Directory for additional information" );

        /** Command string to run the 'dot' tool */
        public final static Property DOT_CALL
                              = new Property( "DotCall", "Dot call pattern" );

        /** Post processing call*/
        public final static Property POST_PROCESSING_CALL
                             = new Property( "PostProcessingCall", "System call for postprocessing" );

        //servlet specific parameters
        /** URL of the rdfs file to process */
        public final static Property RDFS_URL = new Property( "RDFSURL", "URL of source rdfs file" );

        private Property( String id, String descr ) { super( id, descr ); }
    }

    /*
    //replacement patterns for result page generation
    public final static String PICTURE_URL_PATTERN = "%PICTURE_URL%";
    public final static String RDFS_URL_PATTERN = "%RDFS_URL%";
    */

    public void init( ServletConfig servletConfig )
    {
        try
        {
            Config.cfg().addDefaultSettings( new ServletInitParameterReader( servletConfig ) );
        }
        catch( ConfigReader.Exception e )
        {
            RDFSchemaGrapher.debug().error( e );
        }

        readProperties();


        mQueryCache = new QueryCache();
    }

    public void initFromSystemProperties()
    {
        Properties props = System.getProperties();
        try
        {
            Config.cfg().addDefaultSettings( new SystemPropertyReader() );
        }
        catch( ConfigReader.Exception e )
        {
            RDFSchemaGrapher.debug().error( e );
        }

        readProperties();
        mQueryCache = new QueryCache();
    }

    private void readProperties()
    {
        Config cfg = Config.cfg();

        try
        {
            mPictureDir = cfg.getPropertyValue( Property.PICTURE_DIR );
            mPictureURL = cfg.getPropertyValue( Property.PICTURE_URL );
            mDotURL = cfg.getPropertyValue( Property.DOT_URL );
            mRDFSDir = cfg.getPropertyValue( Property.RDFS_DIR );
            mDotDir = cfg.getPropertyValue( Property.DOT_DIR );
            mInfoDir = cfg.getPropertyValue( Property.INFO_DIR );

            mDotCallPattern = cfg.getPropertyValue( Property.DOT_CALL );
            mPostProcessingCallPattern = cfg.getPropertyValue( Property.POST_PROCESSING_CALL );

            //mRawResultHTMLPage = servletConfig.getInitParameter( RAW_RESULT_HTML_PAGE );
        }
        catch( Config.Exception e )
        {
            RDFSchemaGrapher.debug().error( e );
        }

    }

    public void doGet( HttpServletRequest request, HttpServletResponse response )
    {
        processRequest( request, response );
    }

    public void doPost( HttpServletRequest request, HttpServletResponse response )
    {
        processRequest( request, response );
    }

    private void processRequest( HttpServletRequest request, HttpServletResponse response )
    {
        Config config = (Config)Config.cfg().clone();

        BufferHandler messageBuffer = new BufferHandler();
        RDFSchemaGrapher.debug().printMessagesTo( messageBuffer );
        RDFSchemaGrapher.debug().printErrorsTo( messageBuffer );

        String urlStr = null;
        try
        {
            config.addSettings( new ServletRequestReader( request ) );

            String picType = config.getPropertyValue( DotExecutor.Property.PICTURE_TYPE );
            urlStr = config.getPropertyValue( Property.RDFS_URL );
            if (urlStr == null)
            {
                sendErrorMsg( response, "Source URL not set" );
                return;
            }

            //autoadd 'http://"
            urlStr = urlStr.trim();
            if (!urlStr.startsWith( "http://" )/* && !urlStr.startsWith( "ftp://" )*/)
            {
                urlStr = "http://" + urlStr;
            }

            RDFSchemaGrapher.debug().message( "source: " + urlStr );

            URL url = new URL( urlStr );
            URLConnection urlConnection = url.openConnection();
            long lastModifiedMillies = urlConnection.getLastModified();

            DateFormat df = DateFormat.getDateTimeInstance( DateFormat.SHORT,
                                                            DateFormat.MEDIUM,
                                                            Locale.GERMAN );
            String lastModified = df.format( new Date( lastModifiedMillies ) );

            //generate request id
            String requestId = lastModifiedMillies + "-" + config;
            System.err.println( requestId );

            //check cache
            QueryCache.Entry cacheEntry = mQueryCache.getEntry( requestId );
            if (cacheEntry != null)
            {
                RDFSchemaGrapher.debug().message( "picture taken from cache ... " );

                String picURL = cacheEntry.getPictureURL();
                String dotURL = cacheEntry.getDotURL();
                sendResultHTMLPage( response, urlStr, picURL, dotURL, messageBuffer, picType.equalsIgnoreCase( "ps" ) );
                return;
            }

            //not in cache -> generate new picture
            cacheEntry = mQueryCache.createNewEntry( requestId );

            //
            String baseFilename = cacheEntry.getCode() + "-" + lastModifiedMillies;

            System.err.println( "Mapping: " + urlStr + " = " + baseFilename );

            //RDFS-URL auslesen und speichern
            InputStream is = urlConnection.getInputStream();
            InputStreamReader isr = new InputStreamReader( is );
            BufferedReader br = new BufferedReader( isr );

            String rdfsFilename = mRDFSDir + File.separator + baseFilename + ".rdfs";
            FileWriter rdfsw = new FileWriter( rdfsFilename );
            PrintWriter rdfsWriter = new PrintWriter( rdfsw );
            while(true)
            {
                String line = br.readLine();
                if (line == null)
                    break;
                rdfsWriter.println( line );
            }

            //add source url
            rdfsWriter.println( "<!-- URL: " + urlStr + "  last modified: " + lastModified + " -->" );
            rdfsWriter.close();
            br.close();

            //open reader for the rdfs-File
            Reader rdfsFileReader = new FileReader( rdfsFilename );

            //create dot filename
            String dotFilename = mDotDir + File.separator + baseFilename + ".dot";
            Writer dotFileWriter = new FileWriter( dotFilename );

            RDFSchemaGrapher grapher = new RDFSchemaGrapher( rdfsFileReader, dotFileWriter );
            grapher.setConfig( config );
            grapher.setLocalURI( urlStr );
            grapher.run();

            String picFilename = mPictureDir + File.separator + baseFilename + "." + picType;
            String picFileURL = mPictureURL +  "/" + baseFilename + "." + picType;
            String dotFileURL = mDotURL + "/" + baseFilename + ".dot";

            DotExecutor exec = new DotExecutor( mDotCallPattern, dotFilename, picFilename, config );
            boolean ok = exec.execute( System.err );
            if (!ok)
            {
                throw( new Exception( "'dot' execution failure: try another dot version!" ) );
            }

            exec = new DotExecutor( mPostProcessingCallPattern, dotFilename, picFilename, config );
            ok = exec.execute( System.err );


            if (mInfoDir != null)
            {
                //write info file(s)
                String infoFilename = mInfoDir + File.separator + baseFilename + ".txt";
                PrintWriter infoFileWriter = new PrintWriter( new FileWriter( infoFilename ) );
                writeEdgeExtractionInfo( infoFileWriter,
                                         grapher.getExtractorProcessingInfo() );
                infoFileWriter.close();
            }

            sendResultHTMLPage( response, urlStr, picFileURL, dotFileURL, messageBuffer, picType.equalsIgnoreCase( "ps" ) );
            cacheEntry.setPictureURL( picFileURL );
            cacheEntry.setDotURL( dotFileURL );
            mQueryCache.addEntry( cacheEntry );

        }
        catch( MalformedURLException e )
        {
            sendErrorMsg( response, "Illegal source url: " + e.getMessage() );
            return;
        }
        catch( UnknownHostException e )
        {
            sendErrorMsg( response, "Unkown host (" + e.getClass().getName() + "): " + e.getMessage() );
            return;
        }
        catch( IOException e )
        {
            sendErrorMsg( response, "Error while connecting to url '" + urlStr + "' ("
                                      + e.getClass().getName() + "): " + e.getMessage() );
            return;
        }
        catch( Exception e )
        {
            sendErrorMsg( response, "Internal Exception occured (" + e.getClass().getName() + "): " + e.getMessage() );
            e.printStackTrace( System.err );
        }
        finally
        {
            RDFSchemaGrapher.debug().deactivate();
        }
    }

    /**
     *
     */
    void sendResultHTMLPage( HttpServletResponse response,
                             String requestURL, String picFileURL, String dotFileURL,
                             BufferHandler messageBuffer,
                             boolean printImageLink )
    {
        try
        {
            String messages = messageBuffer.toString();

            response.setContentType( HTTP.CONTENT.TEXT_HTML );
            PrintWriter	out = response.getWriter();
            out.println( "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 3.2//EN\">" );
            out.println( "<html>" );

            out.println( "  <head>" );
            out.println( "    <title>FRODO RDFSViz Result</title>" );
            out.println( "  </head>" );

            out.println( "  <body bgcolor=\"#FFFFFF\">" );
            out.println( "    <h1>FRODO RDFSViz Result</h1>" );

            if (printImageLink)
                out.println( "<a href=\"" + picFileURL + "\">click to view the generated image</a>" );
            else
                out.println( "    <img src= \"" + picFileURL + "\">" );

            out.println( "    <br>" );


            if (messages.length() > 0)
            {
                out.println( "<h3>Generated messages:</h3>" );
                out.println( "<code><pre>" );
                out.print( HTML.escape( messages ) );
                out.println( "</pre></code><br>" );
            }

            out.println( "<a href=\"" + dotFileURL + "\">here is the generated dot file</a>" );
            out.println( "<br><hr>" );

            out.println( "    This image was generated with the " );
            out.println( "    <a href=\"http://www.dfki.uni-kl.de/frodo/RDFSViz/\">FRODO RDFSViz tool</a>." );
            out.println( "    <br>" );

            if (!printImageLink)
            {
                out.println( "    Note: if no image is displayed (broken image icon), it simply might be to large for your" );
                out.println( "    browser. You can download it with the context menu (right mouse button)." );
            }

            out.println( " </body>" );

            out.println( "</html>" );
            out.flush();
        }
        catch( Exception ignore )
        {
            System.err.println( ignore );
        }
    }

    public void sendErrorMsg( HttpServletResponse response, String errorMsg )
    {
        try
        {
            response.setContentType( HTTP.CONTENT.TEXT_HTML );
            PrintWriter	out = response.getWriter();
            out.println( "<html>" );
            out.println( "<h1>Error:</h1>" );
            out.println( "<pre>" );
            out.println( errorMsg );
            out.println( "</pre>" );
            out.println( "</html>" );
            out.flush();
        }
        catch( Exception ignore )
        {
            System.err.println( ignore );
        }

    }

    private void writeEdgeExtractionInfo( PrintWriter writer,
                                          EdgeExtractor.ProcessingInfo info )
        throws IOException
    {
        Iterator i = info.getStatementInfos().iterator();
        while( i.hasNext() )
        {
            EdgeExtractor.StatementInfo stmtInfo = (EdgeExtractor.StatementInfo)i.next();
            writer.println( stmtInfo.getStatement() );
        }
    }

    private static QueryCache mQueryCache;
    private static String mPictureDir;
    private static String mPictureURL;
    private static String mDotURL;
    private static String mRDFSDir;
    private static String mDotDir;
    private static String mInfoDir;
    private static String mDotCallPattern;
    private static String mPostProcessingCallPattern;

    //private static String mRawResultHTMLPage;

    // -----------------------------------------------

    public static class Test
    {
        public static void main( String[] args )
        {
            GrapherServlet servlet = new GrapherServlet();

            String grapherPicsURL = args[0];
            String dataStorageDir = args[1];

            System.getProperties().put( Property.PICTURE_DIR.getId(), dataStorageDir + "/pics" );
            System.getProperties().put( Property.PICTURE_URL.getId(), grapherPicsURL );
            System.getProperties().put( Property.RDFS_DIR.getId(), dataStorageDir + "/rdfs" );
            System.getProperties().put( Property.DOT_DIR.getId(), dataStorageDir + "/dot" );
            System.getProperties().put( Property.INFO_DIR.getId(), dataStorageDir + "/info" );
            System.getProperties().put( DotExecutor.Property.FONT_PATH.getId(),
                                            "/usr/openwin/lib/X11/fonts/TrueType" );
            System.getProperties().put( Property.DOT_CALL.getId(),
                                        "/project/java/gv/bin/dot -Tgif " + DotExecutor.SOURCE_FILE_PATTERN
                                        + " -o " + DotExecutor.PICTURE_FILE_PATTERN );
            System.getProperties().put( RDFSchemaGrapher.Property.RDFS_NAMESPACE.getId(), "http://www.w3.org/2000/01/rdf-schema#" );
            servlet.initFromSystemProperties();

            DebugHttpServletRequest debugRequest = new DebugHttpServletRequest();
            debugRequest.parameters.put( RDFSchemaGrapher.Property.PRINT_NAMESPACES.getId(), "on" );
            debugRequest.parameters.put( RDFSchemaGrapher.Property.SHOW_NON_LITERAL_PROPERTIES.getId(), "no" );
            debugRequest.parameters.put( RDFSchemaGrapher.Property.NODE_COLOR.getId(), "green" );
            debugRequest.parameters.put( RDFSchemaGrapher.Property.PROPERTY_EDGE_COLOR.getId(), "red" );

            for( int i=2; i<args.length; ++i )
            {
                debugRequest.parameters.put( Property.RDFS_URL.getId(), args[i] );
                servlet.doGet( debugRequest, new DebugHttpServletResponse() );
            }
        }

    }

}