/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util;

/**
 * Base class for creating typesave enumerated constants.
 * Enumeration ids have an unique id (integer value) and a
 * name. The name, for example, is useful for generation of meanigful
 * messages.<br>
 * <b>Advantages:</b><br>
 *    <ul>
 *      <li>Type checking at compile time.</li>
 *      <li>Fast equality checks with <code>==</code> and <code>!=</code>.</li>
 *      <li>Support for meaningful message generation and reading
 *          enumerated constants from strings.</li>
 *    </ul>
 * Define your own class for each set of enumeration ids. This class
 * should have a private constructor. It has to contain all
 * possible ids as public and final class members.
 * <pre>
 * public class Param extends EnumID
 * {
 *    public final static int _A = 0;
 *    public final static int _B = 1;
 *    public final static Param A = new Param( _A, "A" );
 *    public final static Param B = new Param( _B, "B" );
 *    ...
 *    public final static EnumID[] ALL_PARAMS = new Param[] { A, B };
 *
 *    private Param( int id, String name ) { super( id, name ); }
 * }
 * </pre>
 * Use the newly definied <code>Param</code> class as follows:
 * <pre>
 * //method calls
 * Result result = o.someMethod( Param.A );
 *
 * //reading from a string:
 * String inputStr = readStringFromSomewhere();
 * Param p = Param.parseEnumID( inputStr, Param.ALL_PARAMS );
 *
 * //enumeration checks
 * if (p == Param.B ) { do_something(); }
 * ...
 * switch( p.value() )
 * {
 *   switch _A: do_something();
 *   switch _B: do_something();
 * }
 *
 * //printing messages
 * System.out.println( "The parameter has the value: " + p );
 * </pre>
 */
abstract public class EnumID
{
   /**
    * @param val code of the enumeration id.
    * @param name - name of enumeration id.
    */
   public EnumID( int val, String name )
    {
        mValue = val;
        mName = name;
    }

    /**
     * @return enumeration id code.
     */
    public int value() { return( mValue ); }

    /**
     * @return ID name of accesstype
     */
    public String getName() { return( mName ); }

    /**
     * @return String repressentation of accesstype
     */
    public String toString() { return( getName() ); }

    /**
     * Create an enumeration id from a string source.
     * @param name The string with the expected id name.
     * @param eList List of candidate enumeration ids.
     */
    public static EnumID parseEnumID( String name, EnumID[] eList )
    {
        if (eList == null)
           return( null );

        int index = 0;
        int len = eList.length - 1;

        while(index <= len)
        {
            EnumID id = eList[index];
            if(id.getName().equals( name ))
            {
                return( id );
            }

            index++;
        }

        return( null );
    }

    public static EnumID findEnumID( int value, EnumID[] eList )
    {
        if (eList == null)
           return( null );

        for(int index=0; index<eList.length; ++index)
        {
            if (eList[index].value() == value)
               return( eList[index] );
        }

        //no appropriate EnumID found
        return( null );
    }


    private int mValue;
    private String mName;
}