/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.application;

import java.util.Collection;
import java.util.LinkedList;

/**
 * Represents a command line option.
 * The properties of a command line option is maintaind
 * in this class. The {@link OptionReader} uses this
 * imformation to parse the command line.
 */
public class Option implements Cloneable
{
    public final static boolean UNIQUE = true;
    public final static boolean MULTIPLE = false;

    public final static boolean NEEDS_VALUE = true;
    public final static boolean WITHOUT_VALUE = false;

    public final static boolean NECESSARY = true;
    public final static boolean NOT_NECESSARY = false;

    /*
    public Option( String name, int code, boolean isUnique, String defaultValue )
    {
    }
    */

    public Option( String name, int code, boolean needsValue,
                   boolean isUnique, boolean necessary,
                   String defaultValue )
    {
        this( name, code, needsValue, isUnique, necessary, defaultValue, "" );
    }

    public Option( String name, int code, boolean needsValue,
                   boolean isUnique, boolean necessary,
                   String defaultValue,
                   String description )
    {
        mName = name;
        mCode = code;
        mNeedsValue = needsValue;
        mIsUnique = isUnique;
        mIsNecessary = necessary;
        mValue = null;
        mValues = new LinkedList();
        mDefaultValue = defaultValue;
        mDescription = description;
    }

    public Option( Option other )
    {
        mName = other.mName;
        mCode = other.mCode;
        mNeedsValue = other.mNeedsValue;
        mIsUnique = other.mIsUnique;
        mIsNecessary = other.mIsNecessary;

        mDefaultValue = other.mDefaultValue;
        mValue = other.mValue;
        mValues = (LinkedList)other.mValues.clone();
    }

    public Object clone()
    {
        return( new Option( this ) );
    }

    public String getName()
    {
        return( mName );
    }

    public int getCode()
    {
        return( mCode );
    }

    public boolean isUnique()
    {
        return( mIsUnique );
    }


    public boolean needsValue()
    {
        return( mNeedsValue );
    }

    public void setValue( String value )
    {
        mValue = value;
        if (!mIsUnique)
            mValues.add( value );
    }

    public String getValue()
    {
        if (mValue == null)
            return( mDefaultValue );

        return( mValue );
    }

    public Collection getValues()
    {
        LinkedList l = mValues;
        String value = getValue();
        if (l.isEmpty() && value != null)
        {
            l = new LinkedList();
            l.add( value );
        }

        return( l );
    }

    public boolean valueIsSet()
    {
        if (mIsUnique)
            return( getValue() != null );
        else
            return( getValues().size() != 0 );
    }

    public boolean isNecessary()
    {
        return( mIsNecessary );
    }

    public void setDescription( String d )
    {
        mDescription = d;
    }

    public String getDescription()
    {
        return( mDescription );
    }

    public String getDefaultValue()
    {
        return( mDefaultValue );
    }

    public String toString()
    {
        String s = mName;
        if (mNeedsValue)
            s += ":" + getValue();
        return( mName );
    }


    private String mName;
    private int mCode;
    private boolean mNeedsValue;
    private boolean mIsUnique;
    private boolean mIsNecessary;

    private String mValue;
    private LinkedList mValues;
    private String mDefaultValue;
    private String mDescription;
}
