/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Copyright:    Copyright (c) DFKI GmbH<p>
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.application;

import java.util.*;

public class OptionReader
{
    public static class Exception extends java.lang.Exception
    {
        public Exception( String msg ) { super( msg ); }
    }

    public OptionReader( Option[] options )
    {
        mCurrentIndex = -1;
        mOptionMap = new TreeMap();
        mGivenOptions = new TreeMap();

        addOption( options );
    }


    public void addOption( Option option )
    {
        mOptionMap.put( option.getName(), option );
    }

    public void addOption( Option[] options )
    {
        for( int i=0; i<options.length; ++i )
        {
            addOption( options[i] );
        }
    }

    public void readArgs( String[] args ) throws Exception
    {
        mCurrentIndex = 0;
        if (args.length == 0)
            mCurrentIndex = -1;

        mArgs = args;
        mGivenOptions = new TreeMap();

        while( hasMoreOptions() )
        {
            Option option = nextOption();
        }

        Collection options = mOptionMap.values();
        Iterator i = options.iterator();
        while( i.hasNext() )
        {
            Option option = (Option)i.next();
            if (!option.isNecessary())
                continue;

            int code = option.getCode();
            boolean given = mGivenOptions.get( new Integer( code ) ) != null;

            if (!given)
            {
                if (option.valueIsSet())
                    mGivenOptions.put( new Integer( code ), option );
                else
                    throw( new Exception( "Option '" + option.getName() + "' is missing" ) );
            }
        }
    }

    public boolean hasOption( Option option )
    {
        int code = option.getCode();
        Option o = (Option)mGivenOptions.get( new Integer( code ) );
        return( o != null );
    }

    public String getOptionValue( Option option )
    {
        int code = option.getCode();
        Option o = (Option)mGivenOptions.get( new Integer( code ) );
        if (o == null)
        {
            return( option.getValue() );
        }

        return( o.getValue() );
    }

    public Collection getOptionValues( Option option )
    {
        int code = option.getCode();
        Option o = (Option)mGivenOptions.get( new Integer( code ) );
        if (o == null)
            return( new LinkedList() );

        return( o.getValues() );
    }

    public boolean hasMoreOptions()
    {
        if (getCurrentIndex() == -1)
            return( false );

        return( isOption( getCurrentIndex() ) );
    }

    public Option nextOption() throws Exception
    {
        String optionStr = getCurrentArg();

        Option option = (Option)mOptionMap.get( optionStr );

        if (option == null)
            throw( new Exception( "Unkown option: '" + optionStr + "'" ) );

        Option oldOption = (Option)mGivenOptions.get( new Integer( option.getCode() ) );


        if (oldOption != null)
        {
            if (option.isUnique())
            {
                throw( new Exception( "Not a multiple option: '" + option + "'" ) );
            }

            option = oldOption;
        }

        increaseIndex();
        if (option.needsValue())
        {
            option = (Option)option.clone();
            if (getCurrentIndex() == -1 || isOption( getCurrentIndex() ))
                throw( new Exception( "Missing argument for: '" + optionStr + "'" ) );

            option.setValue( getCurrentArg() );
            increaseIndex();
        }
        mGivenOptions.put( new Integer( option.getCode() ), option );
        return( option );
    }

    private boolean increaseIndex()
    {
        mCurrentIndex++;
        if (mCurrentIndex >= mArgs.length)
            mCurrentIndex = -1;

        return( mCurrentIndex != -1 );
    }

    private int getCurrentIndex()
    {
        return( mCurrentIndex );
    }

    private String getCurrentArg()
    {
        int idx = getCurrentIndex();
        if ( idx == -1)
            return( null );

        return( mArgs[idx] );
    }

    private boolean isOption( int pos )
    {
        return( mArgs[pos].startsWith( "-" ) );
    }

    public String[] pendingArgs()
    {
        String[] pending = null;
        int idx = getCurrentIndex();
        if (idx > -1 && idx < mArgs.length)
        {
            int pl = mArgs.length - idx;
            pending = new String[pl];
            for( int i=0; i<pl; ++i)
            {
                pending[i] = mArgs[idx+i];
            }
        }
        return( pending );
    }

    public String getOverview()
    {
        String s = "Options: \n";
        Iterator i = mOptionMap.values().iterator();
        while( i.hasNext() )
        {
            Option option = (Option)i.next();
            String description = option.getDescription();
            String defaultValue = option.getDefaultValue();
            s += option.getName() + " ";
            if (option.needsValue())
                s += "<val> ";
            if (description.length() != 0)
               s += " : " + description;
            if ( defaultValue != null)
               s += "\n        default: '" + defaultValue + "'";
            s += "\n";
        }
        return( s );
    }

    private int mCurrentIndex;
    private String[] mArgs;
    private TreeMap/*String,Option*/ mOptionMap;
    private TreeMap/*int,boolean*/ mGivenOptions;
}