/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.config;

import java.util.Map;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Properties;

import java.io.Writer;
import java.io.StringWriter;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.IOException;

import dfki.util.Util;
import dfki.util.Strings;

/**
 * Class to support application and software module configuration.
 * An instance of the <code>Config</code> class stores a number
 * of property settings. A property can be either single valued
 * (like the java {@link java.util.Properties}) or it can have multiple
 * values (an array of values). Values are always strings.
 *
 * Settings and default settings are distinguished. They
 * can be set by using the <code>setProperyValue(...)</code>
 * methods or by applying {@link ConfigReader}s.
 * ConfigReaders can read settings from various sources
 * (probably stored in different representations).
 * See direct subclasses of {@link ConfigReader}
 * for currently availabe readers.<br>
 * Additional self made readers can be defined by extending the
 * abstract {@link ConfigReader} class.<br>
 * <br>
 * <b>When property values change:</b>
 * It is possible to subscribe to config changes by adding
 * a {@link Config.ChangeListener}. Whenever the settings
 * of the config intstance to which the listener is subscribed to change,
 * its {@link Config.ChangeListener#hasChanged()} method
 * is called.
 */
public class Config implements Cloneable
{
    public static class Exception extends java.lang.Exception
    {
        public Exception( String msg ) { super( msg  ); }
    }

    /** Exception denoting illegal value/multivalue assignments. */
    public static class MultiValueException extends Exception
    {
        public MultiValueException( Property p )
        {
            super( "'" + p.getId() + "' is " + (p.isMultiValued()? "" : "not") + " a multivalue property"  );
        }
    }

    /**
     * Defines a property in a configuration. Values for
     * a property can be read and set via the {@link Config} api.<br>
     * Properties have a name, a description and a type denoting
     * whether they can store single (string) values or multiple
     * values (arrays of strings).
     */
    public static abstract class Property
    {
        /** Path separator string */
        public final static String SEPARATOR = ".";

        /** Constant for multi valued properties */
        public final static boolean MULTI_VALUE = true;

        /** Constant for single valued properties */
        public final static boolean SINGLE_VALUE = false;


        protected Property( String id, String description )
        {
            this( id, description, false );
        }

        protected Property( String id, String description, boolean isMultiValued )
        {
            mId = id;
            mDescription = description;
            mIsMultiValued = isMultiValued;
        }


        public String getId()
        {
            return( mId );
        }

        public String getDescription()
        {
            return( mDescription );
        }

        public String toString()
        {
            return( mDescription );
        }

        public boolean isMultiValued()
        {
            return( mIsMultiValued );
        }

        private String mId;
        private String mDescription;
        private boolean mIsMultiValued;
    }

    // -----------------------------------------------------

    /**
     * Listener interface for publishing change events.
     */
    public interface ChangeListener
    {
        /**
         * This method is called when some config settings
         * may have changed.
         */
        public void hasChanged();
    }

    // -----------------------------------------------------


    public final static String ENABLED = "on";

    private static Config mConfig = null;

    synchronized public static Config cfg()
    {
        if (mConfig == null)
        {
            mConfig = new Config(/* System.getProperties()*/ );
        }
        return( mConfig );
    }

    synchronized public static Config cfg( ConfigReader r )
        throws ConfigReader.Exception
    {
        Config c = cfg();
        c.addSettings( r );
        return( c );
    }

    private Config()
    {
        mSettings = new TreeMap();
        mDefaultSettings = new TreeMap();
        mChangeListeners = new LinkedList();
    }

    private Config( Properties props )
    {
        this();
        try
        {
            addDefaultSettings( new JavaPropertyReader( props ) );
        }
        catch( ConfigReader.Exception shouldNotHappen )
        {
            Util.debug().error( shouldNotHappen );
        }
    }

    /**
     * Creates a configuration with default values read by
     * the given <code>ConfigReader</code>.
     * If an {@link ConfigReader.Exception} is thrown
     * the <code>Config</code> instance is in an undefined state,
     * i.e. some properties might have been stored others
     * may not.
     */
    private Config( ConfigReader reader )
        throws ConfigReader.Exception
    {
        this();
        addDefaultSettings( reader );
    }

    /**
     * Used by the {@link #clone()} method to obtain a copy of
     * a <code>Config</code> instance.
     */
    private Config( Config other )
    {
        mSettings = new TreeMap( other.mSettings );
        mDefaultSettings = new TreeMap( other.mDefaultSettings );
        mChangeListeners = new LinkedList();
    }

    /**
     * Creates a copy of the configuration.
     */
    public Object clone()
    {
        return( new Config( this ) );
    }


    public void addChangeListener( ChangeListener l )
    {
        mChangeListeners.add( l );
    }

    public void removeChangeListener( ChangeListener l )
    {
        mChangeListeners.remove( l );
    }

    /**
     * Adds some default settings to the current configuration.
     */
    public void addDefaultSettings( ConfigReader r )
        throws ConfigReader.Exception
    {
        fetchMap( mDefaultSettings, r.readPropertySettings() );
        reportChange();
    }

    /**
     * Adds some new settings to the current configuration. Values
     * for a particular property override given default values.
     * @throws ConfigReader.Exception
     */
    public void addSettings( ConfigReader r )
        throws ConfigReader.Exception
    {
        fetchMap( mSettings, r.readPropertySettings() );
        reportChange();
    }

    /**
     * Reads the value for a given property.
     * @param p The given property.
     * @return The value for the property.
     */
    public String getPropertyValue( Property p )
        throws Config.MultiValueException
    {
        return( getPropertyValue( null, p ) );
    }

    /**
     * Reads the value of a given property local to
     * a particular module. The property id has the modul name
     * as a prefix: <code>Modulename.Property = ...</code>
     */
    public String getPropertyValue( final String moduleName, final Property p )
        throws Config.MultiValueException
    {
        if (p.isMultiValued())
           throw( new MultiValueException( p ) );

        String id = p.getId();
        if (moduleName != null)
           id = moduleName + Property.SEPARATOR + id;

        Object value = getPropertyObject( id );

        return( toValue( value ) );
    }

    /**
     * Searches for the value of the given property <code>p</code>.
     * If the property value is not defined this methods searches
     * the value in the enclosing module(s) (from innermost to the outermost).<br>
     * Example: For the Propery <code>Module.SubModule.Debug</code> this
     * method will search the properties <code>Module.SubModule.Debug</code>,
     * <code>Module.Debug</code> and <code>Module.SubModule.Debug</code> (in this
     * order).
     */
    public String getPropertyValueInPath( final Property p )
        throws Config.MultiValueException
    {
        String value = getPropertyValue( p );

        if (value != null)
           return( value );

        //original propety not found. Try path ...
        return( toValue( getPropertyObjectInPath( p.getId() ) ) );
    }


    /**
     * Uses the same search strategy as {@link #getPropertyValueInPath(Config.Property)}.
     * The string <code>moduleName</code> is used as an prefix.
     * @see #getPropertyValueInPath(Config.Property)
     * @see #getPropertyValue(String,Config.Property)
     */
    public String getPropertyValueInPath( final String moduleName, final Property p )
        throws Config.MultiValueException
    {
        String value = getPropertyValue( moduleName, p );

        if (value != null)
           return( value );

        String id = p.getId();
        if (moduleName != null)
           id = moduleName + Property.SEPARATOR + id;

        return( toValue( getPropertyObjectInPath( id ) ) );
    }




    /**
     * Reads the multivalue of a given property.
     */
    public String[] getPropertyValues( Property p )
    {
        String id = p.getId();
        Object value = getPropertyObject( id );

        return( toValues( value ) );
     }

    public String[] getPropertyValues( String moduleName, Property p )
    {
        String id = p.getId();
        if (moduleName != null)
           id = moduleName + Property.SEPARATOR + id;

        Object value = getPropertyObject( id );

        return( toValues( value ) );
    }

    public String[] getDefaultPropertyValues( Property p )
    {
        String id = p.getId();
        Object value = getPropertyObject( id );

        return( toValues( value ) );
    }


    public String[] getPropertyValuesInPath( final Property p )
        throws Config.MultiValueException
    {
        String[] values = getPropertyValues( p );

        if (values != null)
           return( values );

        //original propety not found. Try path ...
        return( toValues( getPropertyObjectInPath( p.getId() ) ) );
    }


    private Object getPropertyObject( final String id )
    {
        Object value = mSettings.get( id );
        if (value == null)
        {
            value = mDefaultSettings.get( id );
       }
       return( value );
    }

    private Object getPropertyObjectInPath( final String propertyId )
        throws Config.MultiValueException
    {
        int dot = propertyId.lastIndexOf( Config.Property.SEPARATOR );
        if (dot == -1)
           return( null );

        String path = propertyId.substring( 0, dot );
        String name = propertyId.substring( dot+1 );

        String newPath = "";
        dot = path.lastIndexOf( Config.Property.SEPARATOR );
        if (dot != -1)
        {
            newPath = path.substring( 0, dot ) + Config.Property.SEPARATOR;
        }

        String id = newPath  + name;
        Object value = getPropertyObject( id );
        if (value == null)
        {
            return( getPropertyObjectInPath( id ) );
        }

        return( value );
    }

    /**
     * Tests if any value has been set for a property.
     * @return <code>true</code>, if a value for the given
     *  property is available (or a default value).
     */
    public boolean hasProperty( Property p )
    {
        return( mSettings.containsKey( p.getId() )
               || mDefaultSettings.containsKey( p.getId() ) );
    }

    public boolean hasDefaultValue( Property p )
    {
        return( mDefaultSettings.containsKey( p.getId() ) );
    }

    /**
     * Checks whether a property is enabled or not.
     * This is helpful for testing checkboxes
     * in HTML-forms when handled by a servlet.
     * @see ServletRequestReader
     * @return <code>true</code>, if the given property is enabled.
     */
    public boolean isEnabled( Property p )
    {
        String[] value = getPropertyValues( p );
        return( value != null && value.length>=1 && value[0].equalsIgnoreCase( ENABLED ) );
    }

    /**
     * Sets the default value for a property.
     */
    public void setDefaultPropertyValue( Property p, String value )
    {
        setValue( mDefaultSettings, p, value );
    }

    /**
     * Sets the default values for a property.
     */
    public void setDefaultPropertyValues( Property p, String[] values )
        throws Config.Exception
    {
        setValues( mDefaultSettings, p, values );
    }

    /**
     * Set the default value for a property <b>only</b> if no
     * value has been set already.
     */
    public void fillInDefaultPropertyValue( Property p, String value )
    {
        if (hasDefaultValue( p ))
           return;

        setDefaultPropertyValue( p, value );
    }

    /**
     * Set the default values for a property <b>only</b> if no
     * values have been set already.
     */
    public void fillInDefaultPropertyValue( Property p, String[] values )
        throws Config.Exception
    {
        if (hasDefaultValue( p ))
           return;

        setDefaultPropertyValues( p, values );
    }

    /**
     * Sets the value for a property, overriding the
     * default value (if any).
     */
    public void setPropertyValue( Property p, String value )
    {
        setValue( mSettings, p, value );
    }


    /**
     * Sets the values for a property, overriding the
     * default value (if any).
     */
    public void setPropertyValues( Property p, String[] values )
        throws Config.Exception
    {
        setValues( mSettings, p, values );
    }

    private void setValue( Map settings, Property p, String value )
    {
        //ignore null values!
        if (value == null)
           return;

        if (p.isMultiValued())
           settings.put( p.getId(), new String[]{ value } );
        else
           settings.put( p.getId(), value );

        reportChange();
    }

    private void setValues( Map settings, Property p, String[] values )
        throws Config.Exception
    {
        //ignore null values!
        if (values == null)
           return;

        if (!p.isMultiValued() && values.length != 1)
           throw( new MultiValueException( p ) );

        mSettings.put( p.getId(), values );

        reportChange();
    }

    public String toString()
    {
        String str = null;
        try
        {
            StringWriter sw = new StringWriter();
            printTo( sw );
            str = sw.toString();
        }
        catch( IOException ignore ) {}

        return( str );
    }


    public void printTo( OutputStream os )
        throws IOException
    {
        printTo( new PrintWriter( os ) );
    }

    public void printTo( Writer w )
        throws IOException
    {
        printTo( new PrintWriter( w ) );
    }

    public void printTo( PrintWriter pw )
        throws IOException
    {
        TreeSet keys = new TreeSet( mSettings.keySet() );
        keys.addAll( mDefaultSettings.keySet() );
        Iterator i = keys.iterator();
        while( i.hasNext() )
        {
            String p = (String)i.next();
            printTo( p, pw );
        }
        pw.flush();
    }


    public String toString( final Property[] props )
    {
        String str = null;
        try
        {
            StringWriter sw = new StringWriter();
            printSettings( props, sw );
            str = sw.toString();
        }
        catch( IOException ignore )
        {
        }

        return( str );
    }

    /**
     * Prints the settigs for the given properties to Writer <code>w</code>.
     */
    public void printSettings( final Property[] props, Writer w )
        throws IOException
    {
        printSettings( props, new PrintWriter( w ) );
    }

    /**
     * Prints the settigs for the given properties to PrintWriter <code>w</code>.
     */
    public void printSettings( final Property[] props, PrintWriter pw )
        throws IOException
    {
        for( int i=0; i<props.length; ++i )
        {
            String id = props[i].getId();
            Object value = getPropertyObject( id );
            pw.println( id + " = " +  toString( value ) );
        }
    }

    private void printTo( String p, PrintWriter pw )
        throws IOException
    {
        Object value = mSettings.get( p );
        pw.print( p + " = " +  toString( value ) );
        String defaultValue = (String)mDefaultSettings.get( p );
        if (defaultValue != null)
        {
            pw.print( "  Default: '" +  defaultValue + "'" );
        }
        pw.println();
    }


    public String printProperty( Property p )
    {
        String str = p.getId() + " = ";
        Object value = getPropertyObject( p.getId() );
        if (value == null)
            str += "<not set>";
        else
            str += toString( value );
        str += "  (" + p.getDescription() + ")";

        return( str );
    }

    private void fetchMap( Map target, Map source )
    {
        if (source == null || target == null)
           return;
        target.putAll( source );
    }

    private String toString( Object o )
    {
        if (o == null)
           return( "<none>" );

        String str = null;
        if (o instanceof String[])
        {
            String[] arg = (String[])o;
            str = Strings.from( arg );
        }
        else
        {
            str = "'" + o.toString() + "'";
        }
        return( str );
    }

    private String[] decode( String values )
    {
        return null;
    }

    private String[] toValues( Object o )
    {
        if (o == null)
           return( null );

        if (o instanceof String)
        {
            return( new String[]{ (String)o } );
        }

        if (o instanceof String[])
        {
            return( (String[])o );
        }

        return( null );
    }

    private String toValue( Object o )
    {
        if (o == null)
           return( null );

        if (o instanceof String)
        {
            return( (String)o );
        }

        if (o instanceof String[])
        {
            String[] vals = (String[])o;
            if (vals.length == 1)
               return( vals[0] );
        }

        return( null );
    }

    /**
     * Inform all registered Listeners of a change.
     */
    private void reportChange()
    {
        Iterator i = mChangeListeners.iterator();
        while( i.hasNext() )
        {
            ((ChangeListener)i.next()).hasChanged();
        }
    }
/*
    public Property getProperty( String s )
    {
        for( int i=0; i<mAllProperties.length; ++i )
        {
            if (mAllProperties[i].getId().equals( s ))
               return( mAllProperties[i] );
        }
        return( null );
    }
*/
    private Map/*String,String|String[]*/ mSettings;
    private Map/*String,String|String[]*/ mDefaultSettings;
    private Collection/*ChangeListener*/ mChangeListeners;
}