
package dfki.util.config;
// Import
import java.util.TreeMap;
import java.util.Map;
import java.util.ArrayList;
import java.util.Iterator;

import java.io.InputStream;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;

import java.util.LinkedList;

// fr DOM/SAX
import org.w3c.dom.*;
import org.xml.sax.*;

// Xerces
import org.apache.xerces.parsers.DOMParser;

public class XMLConfigReader extends ConfigReader
{
    public static class Exception extends ConfigReader.Exception
    {
        public Exception(String message) { super( message ); }
    }

    public static final String LIST_ENRTY_TAG_NAME="val"; // Schlsseltag um Listeneintrge zu kennzeichnen
    // Member
    private InputSource mInput;
    private Document mDocument;
    private TreeMap mConfigMap;

    // Konstruktoren
    public XMLConfigReader ()
    {
        mConfigMap = new TreeMap();
        mInput = new InputSource();
    }

    /**
    Erstellt Konfiguration aus einer Datei
    @param path Pfad der zu lesenden Konfigurationsdatei
    */
    public XMLConfigReader (String path) throws IOException, XMLConfigReader.Exception
    {
        readFromFile(path); // initialisiert und erstellt mInput aus einer Datei
        parseInput(); // initialisiert und erstellt mDocument aus mInput
        parseDocument(); // initialisiert und erstellt mConfigMap aus mDocument
    } // end of XMLConfigReader(String)

    /**
    Generischer Konstruktor zum Lesen einer Konfiguration aus einem InputStream
    @param istream InputStream
    */
    public XMLConfigReader (InputStream istream)
        throws IOException,
               XMLConfigReader.Exception
    {
        readFromInputStream(istream); // initialisiert und erstellt mInput aus einem InputStream
        parseInput(); // initialisiert und erstellt mDocument aus mInput
        parseDocument(); // initialisiert und erstellt mConfigMap aus mDocument
    } // end of XMLConfigReader(InputStream)

    /**
     *Liest Konfiguration von einer URL
     *@param url URL-Objekt der Quelle
    */
    public XMLConfigReader (URL url)
        throws IOException,
               XMLConfigReader.Exception
    {
        readFromURL(url); // initialisiert und erstellt mInput aus URL
        parseInput(); // initialisiert und erstellt mDocument aus mInput
        parseDocument(); // initialisiert und erstellt mConfigMap aus mDocument
    } // end of XMLConfigReader (URL)

    /**
    Generiert mConfigMap aus dem eingelesenen Konfigurationsdokument
    */
    public void parseDocument()
        throws XMLConfigReader.Exception
    {
        mConfigMap= new TreeMap();
        Element rootElement;
        if (mDocument!= null)
        {
            if ((rootElement=mDocument.getDocumentElement())!=null)
            {
                processNodes(rootElement, new ArrayList());
            }
            else
            {
                throw( new Exception("parseDocument:Root Element des Dokuments ist null.") );
            }
        }
        else
        {
            throw( new Exception("parseDokument:Kein Dokument eingelesen bzw. Dokument ist null.") );
        }

    }// end of parseDocument


    /**
     *Erstellt ein Dokument aus dem eingelesenen InputSource Objekt
     */
    public void parseInput()
        throws IOException,
               XMLConfigReader.Exception
    {
        // NonValidatingDOMParser parser = new NonValidatingDOMParser();
        DOMParser parser = new DOMParser();

        if (mInput!=null)
        {
            try
            {
                parser.parse(mInput); // SAXException
                mDocument=parser.getDocument();
            }
            catch (SAXException ex)
            {
                throw( new Exception( "XML-Parse-Fehler:" + ex.getMessage() ) );
            }
        }
        else
        {
            throw( new Exception( "parseInput:InputSource ist null." ) );
        }


    }  // end of parseInput


    private void processNodes(Node currentNode, ArrayList key)
        throws XMLConfigReader.Exception
    {
        // Traversierung des Dokuments (Preorder)
        Node nextNode;
        if (currentNode.getNodeType()==org.w3c.dom.Node.TEXT_NODE)
        {
            // Textknoten gefunden, auf Whitespace berprfen
            String data = (((Text) currentNode).getData()).trim();
            if (data.compareTo("")!=0)
            {
                // Text ist nicht ausschliesslich Whitespace, Mapeintrag
                putToMap (new ArrayList (key), data);
            }
        }

        // Children
        if ((nextNode=currentNode.getFirstChild())!=null)
        {
            if (currentNode.getNodeType()==org.w3c.dom.Node.ELEMENT_NODE)
            // Tagname zum key hinzufgen
            {
                ArrayList newkey= new ArrayList(key);
                newkey.add(((Element) currentNode).getTagName());
                processNodes(nextNode, newkey);
            }
            else
            {
                processNodes(nextNode, key);
            }
         }

        // Sibilings
        if ((nextNode=currentNode.getNextSibling())!=null)
        {
            processNodes(nextNode, key);
        }
        return;
    }// end of processNodes


    public void readFromFile(String path)
        throws IOException,
               XMLConfigReader.Exception
    {
        if (path!=null)
        {
            FileReader file = new FileReader(path);
            mInput = new InputSource(file);
        }
        else
        {
            throw( new Exception("readFromFile:Pfad zur Datei ist null.") );
        }
    }// end of readFromFile

    public void readFromInputStream(InputStream istream)
        throws IOException,
               XMLConfigReader.Exception
    {
        if (istream!=null)
        {
            mInput = new InputSource(istream);
        }
        else
        {
            throw( new Exception("readFromInputStream: oeffnen fehlgeschlagen.") );
        }
    }// end of readFromInputStream


    public void readFromURL(URL url)
        throws IOException,
               XMLConfigReader.Exception
    {
        if (url!=null)
        {
            mInput = new InputSource (url.openStream());
        }
        else
        {
            throw( new Exception("readFromURL:URL-Objekt ist null.") );
        }
    } // end of readFromURL

    private void putToMap(ArrayList key, String data)
        throws XMLConfigReader.Exception
    {
        String keyString;

        if ((key.size()>1) && (key.lastIndexOf(LIST_ENRTY_TAG_NAME)==(key.size()-1)))
        // Listeneintrag gefunden
        {
            Object oldEntries;
            key.remove(key.size()-1); // um LIST_ENRTY_TAG_NAME zu zu entfernen
            keyString = toKeyString(key);
            oldEntries= mConfigMap.get(keyString);

            if (oldEntries instanceof String[]) // Listenelement hinzufgen
            {
                String oldValues[] = (String[]) oldEntries;
                String newValues[] = new String[oldValues.length+1];
                for (int i=0; i<oldValues.length; i++)
                {
                    newValues[i]= oldValues[i];
                }
                newValues[oldValues.length]= data;
                mConfigMap.put(keyString, newValues);
            }
            else if (oldEntries==null) // neue Liste
            {
                mConfigMap.put(keyString, new String[]{data});
            }

            else
            {
                throw( new Exception("putToMap:ungltiger Typ im Map-Eintrag.") );
            }
        }
        else if ((keyString = toKeyString(key))!=null)
        {
            mConfigMap.put(keyString, data);
        }
        else
        {
            // TEXT_NODE gefunden ohne sinnvollen Schlssel, Exception oder ignorieren ?
        }
    }// end of putToMap

    private String toKeyString (ArrayList key)
    // wandelt Liste von Schlssel in die String-Reprsentation um, liefert null falls
    // Liste leer ist oder nur das Root-Element enthlt
    {
        String keyString="";
        if ((key!=null) && (key.size()>1))
        {
            for (int i=1; i<key.size();i=i+1)
            // Fall i==0 uberspringen da root-Element nicht im Schlssel bercksichtigt wird
            {
                if (i>1)
                {
                    keyString=keyString + dfki.util.config.Config.Property.SEPARATOR + (String) key.get(i);
                }
                else if (i==1)
                {
                    keyString= (String) key.get(i);
                }
            }
            return keyString;
        }
        else
        {
            return null;
        }
    }

    public Map readPropertySettings()
    {
        return mConfigMap;
    }


    public static class Test
    {
        public static void main(String[] args)
            throws IOException,
                   XMLConfigReader.Exception
        {
            TreeMap result;
            XMLConfigReader reader;
            Iterator it;
            String currentKey, currentStringValue;
            System.out.println("Test XMLConfigReader Start:");

            for (int i=0; i<args.length; i++)
            {
                String filepath= args[i];// "/home/kaestner/frodo/java/xmlconfig/testfile.xml";
                reader = new XMLConfigReader(filepath);
                result = new TreeMap(reader.readPropertySettings());
                it = (result.keySet()).iterator();
                System.out.println("\nFile: " + filepath);
                while (it.hasNext())
                {
                    currentKey = (String) it.next();
                    if (result.get(currentKey) instanceof String[])
                    {
                        currentStringValue ="[ ";
                        String values[]= (String[]) result.get(currentKey);
                        for (int j=0; j<values.length; j++)
                        {
                            currentStringValue=currentStringValue + values[j]+ " ";
                        }
                        currentStringValue = currentStringValue + "]";
                    }
                    else
                    {
                        currentStringValue= (String) result.get(currentKey);
                    }
                    System.out.println(currentKey+": "+currentStringValue);
                }
            }
            System.out.println("\nTest fertig.");
        }
    }
} // end of class XMLConfigReader