/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.debug;

import dfki.util.EnumID;
import dfki.util.config.Config;
//import dfki.util.config.ConfigReader;

import java.util.Hashtable;
import java.io.*;

/**
 * Provides support for issuing debugging messages.
 *
 * A module, i.e. a set of classes usually grouped together in
 * packages and subpackages, can request it's own instance of
 * the Debug class. The verbosity of output is controlled by
 * switching to certain debug modes
 * (see methods {@link #setDebugMode(Debug.Mode) setDebugMode(...)}).<br>
 * The following messages types are supported:
 * <ul>
 *   <li>{@link Debug.Mode#INFORM_CODE INFORM}: Messages issued to keep the
 *                user informed. For example, showing what action is currently
 *                in progress or printing some (intermediate) results.</li>
 *   <li>{@link Debug.Mode#WARNING_CODE WARNING}: A message denoting a possible
 *              error or some other noteworthy situation.</li>
 *   <li>{@link Debug.Mode#ERROR_CODE ERROR}: A failure which could not properly
 *              handled.</li>
 *   <li>{@link Debug.Mode#TRACE_CODE TRACE}: Messages for tracing the path of execution,
 *              i.e. entering or leaving certain methods.</li>
 * </ul>
 * <br>
 * The following debug modes are available:
 * <ul>
 *   <li>{@link Debug.Mode#NORMAL}</li>
 *   <li>{@link Debug.Mode#WARNING}</li>
 *   <li>{@link Debug.Mode#ERROR}</li>
 *   <li>{@link Debug.Mode#TRACE}</li>
 *   <li>{@link Debug.Mode#FULL}</li>
 *   <li>{@link Debug.Mode#NODEBUG}</li>
 * </ul>
 * <br>
 * One instances of the debugging class is maintained for
 * each module. Use the {@link #forModule(String)} method to
 * obtain it:
 * <pre>
 *   Debug d = Debug.forModule( "ModuleName" );
 *   d.warning( "some warning message" );
 * </pre>
 *
 * The following methods ar available for printing out messages:
 * <pre>
 * Debug d = Debug.forModule( "ModuleName" );
 * d.setDebugMode( Debug.Mode.FULL );
 *
 * //informative messages
 * d.message( "a short message" );
 *
 * //error messages
 * d.error( "a brief error message" );
 * d.error( new Exception() );
 * d.error( this, "error message );
 * d.error( ThisClass.class, "error message" );
 *
 * //warning messages
 * d.warning( "a little warning" );
 * d.warning( this, "a little warning" );
 * d.warning( ThisClass.class, "a little warning" );
 *
 * //trace messages
 * d.entering( this, "printShortMessages()" );
 * d.leaving( this, "printShortMessages()" );
 * d.entering( ThisClass.class, "printShortMessages()" );
 * d.leaving( ThisClass.class, "printShortMessages()" );
 * d.entering( "printShortMessages()" );
 * d.leaving( "printShortMessages()" );
 * </pre>
 * If cofiguration settings are given, the <code>Debug</code> class
 * ties to read the debug mode from them. When the settings change
 * the debug mode is automatically set according to the new settings.<br>
 * The following settings are checked:
 * <ul>
 *   <li><code>Debug.ModuleName</code>: The debug mode for module
 *                    <code>ModuleName</code>. Overrides global or default settings.
 *                     </li>
 *   <li><code>Debug</code>: The global debug mode. Used, if no module specific
 *                     settings are found.</li>
 * </ul>
 *
 * The class {@link dfki.util.examples.debug.DeBugged} shows how the
 * <code>Debug</code> class has to be used.<br>
 */
public class Debug
{
    public static class Exception extends java.lang.Exception
    {
        public Exception( String msg ) { super( msg  ); }
        public Exception( java.lang.Exception e ) { super( e.getMessage() ); }
    }

    public static final class Property extends Config.Property
    {
        public final static Property Debug = new Property( "Debug", "Common debug mode" );

        private Property( String id, String descr )
        {
            super( id, descr );
        }
    }

    public static final class Mode extends EnumID
    {
        /** Code for printing no message at all. */
        public final static int NONE_CODE = 0;

        /** Code for normal messages */
        public final static int INFORM_CODE = 1;

        /** Code for warnings */
        public final static int WARNING_CODE = 2;

        /** Code for errors*/
        public final static int ERROR_CODE = 4;

        /** Code for trace messages */
        public final static int TRACE_CODE = 8;

        /** Code for internal messages */
        public final static int INTERNAL_CODE = 16;

        /**
         * Only error messages should be printed.
         */
        public final static Mode ERROR = new Mode( ERROR_CODE , "ERROR" );
        /**
         * Error and warning messages should be printed.
         */
        public final static Mode WARNING = new Mode( ERROR_CODE
                                                     | WARNING_CODE
                                                     | INTERNAL_CODE, "WARNING" );
        /**
         * Prints standard output messages, error and warning messages.
         */
        public final static Mode NORMAL = new Mode( INFORM_CODE
                                                    | WARNING_CODE
                                                    | ERROR_CODE , "NORMAL" );
        /**
         * Only trace messages should be printed (when a thread enters or leaves a method).
         */
        public final static Mode TRACE = new Mode( TRACE_CODE
                                                   | ERROR_CODE
                                                   | INTERNAL_CODE, "TRACE" );
        /**
         * All messages should be printed.
         */
        public final static Mode FULL = new Mode( INFORM_CODE
                                                  | WARNING_CODE
                                                  | ERROR_CODE
                                                  | TRACE_CODE
                                                  | INTERNAL_CODE, "FULL");
        /**
         * No messages at all should be printed.
         */
        public final static Mode NODEBUG = new Mode( NONE_CODE, "NODEBUG" );

        /** List of all supported debug modes */
        private final static EnumID[] ALL_MODES = { FULL, ERROR, WARNING, TRACE, NORMAL, NODEBUG };

        /**
        * Constructs a debug mode id with a name and a value.
        * @param name The string representation of this mode. The name
        *             is used to parser debug modes from a file or string.
        * @param code The value of this mode holds the information on
        *             what types of messages should be printed out.
        */
        private Mode( int code, String name )
        {
            super( code, name );
        }
    }

    public interface MessageHandler
    {
        public void println( String line );
        public void flush();
        //public void close();
    }

    /** Default mode: {@link Mode#NORMAL} */
    public final static Mode DEFAULT_DEBUG_MODE = Mode.NORMAL;


    /**
     * Sets the debug mode according to the passed string.
     * @param mode Debug mode as a string
     * @throws Debug.Exception In case an <code>mode</code> is an illegal identifer.
     */
    public void setDebugMode( String mode )
        throws Debug.Exception
    {
        setDebugMode( debugLevelFromString( mode ) );
    }

    /**
     * Sets the current debug level according to the passed mode.
     * @param m Debug mode.
     * @see Debug.Mode#NODEBUG
     * @see Debug.Mode#WARNING
     * @see Debug.Mode#ERROR
     * @see Debug.Mode#TRACE
     * @see Debug.Mode#FULL
     */
    public void setDebugMode( Debug.Mode m )
    {
        setDebugMode( m.value() );
    }

    /**
     * Sets the current debug level according to the passed value.
     * @param mode Customized debug mode.
     * @see Debug.Mode#INFORM_CODE
     * @see Debug.Mode#WARNING_CODE
     * @see Debug.Mode#ERROR_CODE
     * @see Debug.Mode#TRACE_CODE
     */
    public void setDebugMode( int mode )
    {
        if (mDebugMode == mode)
           return;

        mDebugMode = mode;
        EnumID lvl = EnumID.findEnumID( mode, Mode.ALL_MODES );
        String id = (lvl != null)?
                         lvl.getName()
                        : ""+mode;
        internal( "Debug mode set to " + id );
    }

    /**
     * Returns the actual DebugLevel by evaluating the passed string
     * @param debugLevel DebugLevel as a string
     * @return DebugLevel corresponding to the passed string
     */
    private int debugLevelFromString( String debugLevel )
        throws Debug.Exception
    {
        EnumID dt = EnumID.parseEnumID( debugLevel, Mode.ALL_MODES );

        if(dt != null)
              return( dt.value() );

        Integer bitmuster = new Integer(0);
        try
        {
            bitmuster = new Integer( Integer.parseInt( debugLevel,2 ) );
            if( bitmuster.intValue() > Mode.FULL.value())
            {
                throw new NumberFormatException(  );
            }
        }
        catch( NumberFormatException e )
        {
            throw( new Exception( "Invalid debug mode: " + debugLevel ) );
        }

        return( bitmuster.intValue() );
    }


    /**
     * @return Current debug mode.
     */
    public int getDebugMode()
    {
      return( mDebugMode );
    }

    private static Hashtable mDebugEntries = new Hashtable();
    //private static Debug mDebug = null;

    /*
    private Debug( String module, Config config )
    {
       mModuleName = module;
       printMessagesTo( System.out );
       printErrorsTo( System.err );
       setConfig( config );
    }
    */

    private Debug( String module )
    {
        mModuleName = module;
        printMessagesTo( System.out );
        printErrorsTo( System.err );
        mConfig = Config.cfg();

        Config.ChangeListener l =
           new Config.ChangeListener()
           {
               public void hasChanged() { updateSettings(); }
           };
        mConfig.addChangeListener( l );
        mDebugMode = DEFAULT_DEBUG_MODE.value();
        updateSettings();
    }

    /**
     * Obtain the debugging instance for the given module and
     * apply <code>config</code> to it.
     * @see dfki.util.config.Config
     *
    public static Debug forModule( String module, Config config )
    {
        Debug out = null;

        if (mDebugEntries.containsKey( module ))
        {
            out = (Debug)mDebugEntries.get( module );
            out.updateSettings();
        }
        else
        {
            out = new Debug( module, config );
            mDebugEntries.put( module, out );
        }

        return ( out );
    }
    */

    /**
     * Obtain the debugging instance for the given module.
     */
    public static Debug forModule( String module )
    {
        Debug out = null;

        if (mDebugEntries.containsKey( module ))
        {
            out = (Debug)mDebugEntries.get( module );
            //out.updateSettings();
        }
        else
        {
            out = new Debug( module );
            mDebugEntries.put( module, out );
        }

        return ( out );
    }

    public static Debug forThread( String module )
    {
        String moduleName = module + Config.Property.SEPARATOR
                                  + Thread.currentThread().getName();
        return( forModule( moduleName ) );
    }

    public void deactivate()
    {
        mDebugEntries.remove( this );
    }

    /**
     * Applies a new configuration.
     *
    public void setConfig( Config config )
    {
        if( config == null )
        {
            config = new Config();
        }
        mConfig = config;

        //ChangeListener eintragen
        Config.ChangeListener l =
           new Config.ChangeListener()
           {
               public void hasChanged() { updateSettings(); }
           };
        config.addChangeListener( l );
        updateSettings();
    }
    */

    /**
     * Checks the relevant properties for the <code>Debug<code> class
     * and changes debug mode, if neccessary.
     */
    public void updateSettings()
    {
        try
        {
            String dLev = mConfig.getPropertyValueInPath( mModuleName, Debug.Property.Debug );
            if (dLev == null)
            {
                return;
            }

            setDebugMode( dLev );
        }
        catch( Config.Exception shouldNotHappen )
        {
            error( shouldNotHappen );
        }
        catch( Exception e )
        {
            error( e );
            setDebugMode( DEFAULT_DEBUG_MODE );
        }
    }

    /**
     * Redirects messages to the given PrintWriter.
     */
    public void printMessagesTo( PrintWriter pw )
    {
        mOutHandler = new PrintWriterAdapter( pw );
        //mOutWriter = pw;
    }

    /**
     * Redirects messages to the given OutputStream.
     */
    public void printMessagesTo( OutputStream os )
    {
        printMessagesTo( new PrintWriter( os ) );
    }

    /**
     * Redirects messages to the given MessageHandler.
     */
    public void printMessagesTo( MessageHandler h )
    {
        mOutHandler = h;
    }

    /**
     * Redirects error messages to the given PrintWriter.
     */
    public void printErrorsTo( PrintWriter pw )
    {
        mErrHandler = new PrintWriterAdapter( pw );
    }

    /**
     * Redirects error messages to the given OutputStream.
     */
    public void printErrorsTo( OutputStream os )
    {
        printErrorsTo( new PrintWriter( os ) );
    }

    /**
     * Redirects error messages to the given MessageHandler.
     */
    public void printErrorsTo( MessageHandler h )
    {
        mErrHandler = h;
    }

    //------

    public void message( String line )
    {
        if ((Mode.INFORM_CODE & mDebugMode) !=0)
        {
           out().println( prettyModuleName() + ">" + line);
           out().flush();
        }
    }

    public void message( Object o, String line )
    {
        if ((Mode.INFORM_CODE & mDebugMode) !=0)
        {
           out().println( prettyModuleName() + "> " + o.getClass().getName() + ": " + line);
           out().flush();
        }
    }

    public void message( Class c, String line )
    {
        if ((Mode.INFORM_CODE & mDebugMode) !=0)
        {
           out().println( prettyModuleName() + "> " + c.getName() + ": " + line);
           out().flush();
        }
    }

    public void warning( String line )
    {
        if ((Mode.WARNING_CODE & mDebugMode)!=0)
        {
           err().println( prettyModuleName() + "Warning: " + line );
           err().flush();
        }
    }

    public void warning( Throwable e )
    {
        warning( e.toString() );
    }

    public void warning( Object o, String line )
    {
        if ((Mode.WARNING_CODE & mDebugMode) !=0)
        {
           err().println( prettyModuleName() + "Warning: " + o.getClass().getName() + ": " + line);
           err().flush();
        }
    }

    public void warning( Class c, String line )
    {
        if ((Mode.WARNING_CODE & mDebugMode) !=0)
        {
           err().println( prettyModuleName() + "Warning: " + c.getName() + ": " + line);
           err().flush();
        }
    }

    /**
     * Prints the exception message and the stack strace to the error stream.
     */
    public void error( Throwable exc )
    {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter( sw );

        exc.printStackTrace( pw );
        if ((Mode.ERROR_CODE & mDebugMode)!=0)
        {
           err().println( prettyModuleName() + " Error: " + sw.toString());
           err().flush();
        }
    }


    public void error( String msg )
    {
        if ((Mode.ERROR_CODE & mDebugMode)!=0)
        {
           err().println( prettyModuleName() + " Error: " + msg );
           err().flush();
        }
    }

    public void error( Object o, String line )
    {
        if ((Mode.ERROR_CODE & mDebugMode)!=0)
        {
           err().println( prettyModuleName() + " Error: " + o.getClass().getName() + "." + line );
           err().flush();
        }
    }

    /**
     * Error messages issued from static methods
     * (where no <code>this</code> object exists).
     */
    public void error( Class c, String line )
    {
        if ((Mode.ERROR_CODE & mDebugMode)!=0)
        {
           err().println( prettyModuleName() + " Error: " + c.getName() + "." + line );
           err().flush();
        }
    }

    /**
     * Trace message indicating the start of a method.
     */
    public void entering( String line )
    {
        if ((Mode.TRACE_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + " entering: " + line );
           out().flush();
        }
    }

    /**
     * Trace message indicating the start of a method.
     */
    public void entering( Object o, String line )
    {
        if ((Mode.TRACE_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + " entering: " + o.getClass().getName() + "." + line );
           out().flush();
        }
    }

    /**
     * Trace message indicating the start of a method.
     */
    public void entering( Class c, String line )
    {
        if ((Mode.TRACE_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + " entering: " + c.getName() + "." + line );
           out().flush();
        }
    }

    /**
     * Trace message indicating the completion of a method.
     */
    public void leaving( String line )
    {
        if ((Mode.TRACE_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + " leaving: " + line );
           out().flush();
        }
    }

    /**
     * Trace message indicating the completion of a method.
     * The object <code>o</code> denotes the instance the
     * method belongs to.
     */
    public void leaving( Object o, String line )
    {
        if ((Mode.TRACE_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + " leaving: " + o.getClass().getName() + "." + line );
           out().flush();
        }
    }

    /**
     * Trace message indicating the completion of a method.
     * This method is to be used in static methods, where
     * no <code>this</code> object exists.
     */
    public void leaving( Class c, String line )
    {
        if ((Mode.TRACE_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + " leaving: " + c.getName() + "." + line );
           out().flush();
        }
    }


    /**
     * Prints out an internal message, for example when the debugging mode
     * has been changed.
     */
    public void internal( String line )
    {
        if ((Mode.INTERNAL_CODE & mDebugMode)!=0)
        {
           out().println( prettyModuleName() + ":" + line);
           out().flush();
        }
    }



    private String prettyModuleName()
    {
        String mn = mModuleName;
        if (mn != null && mn.length()>0)
        {
            mn = "[" + mn + "]";
        }
        else
        {
            mn = "";
        }
        return( mn );
    }

    private MessageHandler out()
    {
        return( mOutHandler );
    }

    private MessageHandler err()
    {
        return( mErrHandler );
    }

    /** The current debug level */
    private int mDebugMode = -1;

    private String mModuleName;

    private MessageHandler mOutHandler = null;
    private MessageHandler mErrHandler = null;

    private Config mConfig;
}