/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.exec;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.PrintStream;
import java.io.IOException;

import dfki.util.Util;

/**
 * Class for executing an external application.
 */
public class Executor
{
    /**
     * Class for copying text lines from an InputStream to a
     * given PrintStream.
     */
    public static class StreamGobbler extends Thread
    {
        public StreamGobbler( InputStream is, String type )
        {
            this( is, type, null );
        }

        public StreamGobbler( InputStream is, String type, PrintStream ps )
        {
            mType = type;
            mInputStream = is;
            mOutputStream = ps;
        }

        public void run()
        {
            try
            {
                InputStreamReader isr = new InputStreamReader( mInputStream );
                BufferedReader br = new BufferedReader( isr );

                while(true)
                {
                    String line = br.readLine();
                    if (line == null)
                        break;
                    if (mOutputStream != null)
                    {
                       mOutputStream.println( mType + line );
                       mOutputStream.flush();
                    }
                }
                //br.close();
            }
            catch( IOException e )
            {
                Util.debug().warning( e );
            }
        }

        private String mType;
        private InputStream mInputStream;
        private PrintStream mOutputStream;
    }

    public Executor( String toExecute )
    {
        this( toExecute, null );
    }

    public Executor( String toExecute, String[] env )
    {
        mToExecute = toExecute;
        mToExecuteArr = null;
        mEnvironment = env;
    }

    public Executor( String[] execArr, String[] env )
    {
        mToExecute = null;
        mToExecuteArr = execArr;
        mEnvironment = env;
    }

    public boolean execute( PrintStream ps )
    {
        boolean ok = true;
        Runtime runtime = Runtime.getRuntime();

        if ((mToExecute == null || mToExecute.length() == 0)
             && (mToExecuteArr == null || mToExecuteArr.length == 0 ))
        {
            return( true );
        }

        try
        {
            Process p = null;
            if (mToExecute != null)
               p = runtime.exec( mToExecute, mEnvironment );
            else
               p = runtime.exec( mToExecuteArr, mEnvironment );

            StreamGobbler outGobbler = new StreamGobbler( p.getInputStream(), "> ", ps );
            StreamGobbler errGobbler = new StreamGobbler( p.getErrorStream(), "Error: ", ps );

            outGobbler.start();
            errGobbler.start();

            p.waitFor();

            outGobbler.join();
            errGobbler.join();

            ok = (p.exitValue() == 0);
        }
        catch( Exception e )
        {
            ok = false;
            ps.println( e );
        }
        finally
        {
            if (!ok)
               ps.println( "execution not successful !" );
        }
        return( ok );
    }

    private String mToExecute;
    private String[] mToExecuteArr;
    private String[] mEnvironment;
}
