/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Copyright:    Copyright (c) DFKI GmbH<p>
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.io;

import java.io.File;
import java.io.StringWriter;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.RandomAccessFile;
import java.io.IOException;

import dfki.util.Util;

public class Files
{
    /**
     * Maximum size (in bytes) of the buffer used to copy files.
     * @see #copy(String,String)
     */
    public final static int MAX_COPY_BUFFER_SIZE = 2048;

    public static String currentWorkingDir()
    {
        return( (String)System.getProperties().get( "user.dir" ) );
    }

    public static String getAbsoluteFilename( String relFilename )
    {
        if (relFilename == null)
           return( null );

        relFilename = relFilename.trim();
        return( new File( relFilename ).getAbsolutePath() );
    }

    public static File getAbsoluteFile( final String relFilename )
    {
        if (relFilename == null)
           return( null );

        return( new File( relFilename ).getAbsoluteFile() );
    }


    public static String readerToString( final Reader reader )
        throws IOException
    {
        BufferedReader br = new BufferedReader( reader );

        StringWriter sr = new StringWriter();
        PrintWriter pw = new PrintWriter( sr );

        String line = null;
        while( true )
        {
            line = br.readLine();
            if (line == null)
               break;
            pw.println( line );
        }
        pw.close();
        return( sr.getBuffer().toString() );
    }

    /**
     * Returns the content of a given file as a string.
     */
    public static String fileToString( final String fileName )
        throws IOException
    {
        FileReader fr = new FileReader( getAbsoluteFilename( fileName ) );
        return( readerToString( fr ) );
    }


    /**
     * Copies a file to another.
     */
    static void copy( final String source, final String target )
        throws IOException
    {
        copy( new File( source), new File( target ) );
    }

    /**
     * Copies a file to another.
     */
    static void copy( final File source, final File target )
        throws IOException
    {
      Util.debug().entering( Files.class, "copy(String, String)" );

      try
      {
          RandomAccessFile ra = new RandomAccessFile(source,"r");
          RandomAccessFile rw = new RandomAccessFile(target,"rw");

          long size = ra.length();
          long loops = size/getMaxCopyBufferSize();
          long restsize = size - loops*MAX_COPY_BUFFER_SIZE;
          byte buffer[] = new byte[MAX_COPY_BUFFER_SIZE];
          byte restBuff[] = new byte[(int)restsize];

          //int data = 0;
          //long pointer = ra.getFilePointer();

          Util.debug().message("Copying File: " + source + " to " + target);

          for( int i = 0; i< loops; i++)
          {
            ra.readFully(buffer);
            rw.write(buffer);
          }

          ra.readFully(restBuff);
          rw.write(restBuff);

          //System.out.println();

          ra.close();
          rw.close();

          if ( (size / 1024) <= 0 )
          {
            Util.debug().message(size + " Bytes copied.");
          }
          else if ( ((size / 1024) >= 1) || ((size / 1024) < 1024) )
          {
            Util.debug().message(size/1024 + " KBytes copied.");
          }
          else
          {
            Util.debug().message(size/1024/1024 + " MBytes copied.");
          }
      }
      finally
      {
        Util.debug().leaving( Files.class, "copy(String, String)" );
      }
    }

    private static int getMaxCopyBufferSize()
    {
        return( MAX_COPY_BUFFER_SIZE );
    }
}