/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.rdf;

import java.io.Reader;
import java.io.Writer;
import java.io.PrintStream;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.util.Map;
import java.util.Collection;
import java.util.TreeMap;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import dfki.util.config.Config;
import dfki.util.reflect.Objects;
import dfki.util.Util;

import dfki.util.xml.XML;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import java.net.URL;
import java.net.MalformedURLException;

import org.w3c.rdf.model.Model;
import org.w3c.rdf.model.ModelException;
import org.w3c.rdf.model.Resource;
import org.w3c.rdf.model.Statement;
import org.w3c.rdf.model.Literal;
import org.w3c.rdf.model.RDFNode;

import org.w3c.rdf.syntax.RDFParser;
import org.w3c.rdf.syntax.RDFSerializer;
import org.w3c.rdf.syntax.SerializationException;

import org.w3c.rdf.util.RDFFactory;
import org.w3c.rdf.model.NodeFactory;
import org.w3c.rdf.util.SetOperations;

import org.xml.sax.SAXException;
import org.xml.sax.InputSource;

import org.w3c.rdf.util.ModelConsumer;

/**
 * Class to support RDF handling.
 * Main entry points are the two methods for creating an {@link org.w3c.rdf.util.RDFFactory}.
 * These are {@link #factory()} and {@link #factory(String)}. New models, RDF parsers and
 * RDF serializers can be created via this factory class.<br>
 *
 * The RDF class also provides a number of static convenience methods that
 * make life with RDF a little bit easier.<br>
 *
 * A speacial {@link RDF.Syntax} class offers standard resource instances
 * for given namespaces. These instances can be useful for search and
 * compare operations on the triples of a model. The <code>Syntax</code> instance
 * can be obtained via the methods {@link #syntax()} and
 * {@link #syntax(String)}.<br>
 *
 * <br>
 * Properties of the <code>RDF</code> class are:
 * <dl>
 * <dt><b>RDF_FACTORY_IMPL_CLASSNAME:</b> (default: <code>Util.RDF.RDFFactoryImplClassName</code>)</dt>
 *       <dd>Name of the class implementing the {@link org.w3c.rdf.util.RDFFactory} interface</dd>
 * <dt><b>DEFAULT_SYNTAX_NAMESPACE:</b> (default: <code>Util.RDF.DefaultSyntaxNamespace</code>)</dt>
 *       <dd>Default namespace for identifiers that are related to the RDF specification.</dd>
 *</dl>
 * <br>
 * Have a look at {@link dfki.util.examples.rdf} to see how to use
 * the <code>RDF</code> class.
 */
public final class RDF
{
    public final static String MODULE_NAME = "RDF";

    public final static String DEFAULT_RDF_FACTORY_CLASSNAME = "org.w3c.rdf.util.RDFFactoryImpl";
    public final static String DEFAULT_SYNTAX_NAMESPACE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
    public final static String DEFAULT_IMPORT_IMPL_CLASSNAME = "dfki.util.rdf.import_5_7_01.RDFImportImpl";

    static
    {
        Setup.setDefaults();
    }

    public static class Exception extends java.lang.Exception
    {
        public Exception( String msg ) { super( msg  ); }
    }


    /**
     * Defines the Properties of the {@link RDF} class.
     */
    public static class Property extends dfki.util.Util.Property
    {
        private Property( String path, String descr )
        {
            super( MODULE_NAME + Config.Property.SEPARATOR + path, descr );
        }

        /** Class name of the RDF factory implementation. */
        public final static Property RDF_FACTORY_IMPL_CLASSNAME
                              = new Property( "RDFFactoryImplClassName",
                                              "Class name of the RDF factory implementation" );
        /** Default namespace for rdf resources. */
        public final static Property DEFAULT_SYNTAX_NAMESPACE
                              = new Property( "DefaultSyntaxNamespace",
                                              "Default namespace for rdf resources" );

        public final static Property IMPORT_IMPL_CLASSNAME
                              = new Property( "ImportImplClassName",
                                              "Class name of the rdf import implementation" );

        protected final static Property[] ALL_PROPERTIES = { RDF_FACTORY_IMPL_CLASSNAME,
                                                             DEFAULT_SYNTAX_NAMESPACE,
                                                             IMPORT_IMPL_CLASSNAME };
    }


    /**
     * Class managing the Setup of the {@link RDF} class.
     */
    public static class Setup
    {
        static
        {
            try
            {
                Class.forName( XML.class.getName() );
                Setup.setDefaults();
            }
            catch( java.lang.Exception e )
            {
                Util.debug().error( e );
            }
        }

        public static void setDefaults()
        {
            Util.config().setDefaultPropertyValue( Property.RDF_FACTORY_IMPL_CLASSNAME, DEFAULT_RDF_FACTORY_CLASSNAME );
            Util.config().setDefaultPropertyValue( Property.DEFAULT_SYNTAX_NAMESPACE, RDF.DEFAULT_SYNTAX_NAMESPACE );
            Util.config().setDefaultPropertyValue( Property.IMPORT_IMPL_CLASSNAME, RDF.DEFAULT_IMPORT_IMPL_CLASSNAME );
        }

        public static void setRDFFactoryImplClassName( String className )
        {
            Util.config().setPropertyValue( Property.RDF_FACTORY_IMPL_CLASSNAME, className );
        }
    }

    public static class Syntax
    {
        private Syntax( String namespace )
            throws RDF.Exception
        {
            mNodeFactory = factory().getNodeFactory();

            mNamespace = namespace;
            try
            {
                mObject = createResource( "object" );
                mStatement = createResource( "Statement" );
                mAlt = createResource( "Alt" );
                mSubject = createResource( "subject" );
                mValue = createResource( "value" );
                mPredicate = createResource( "predicate" );
                mProperty = createResource( "Property" );
                mSeq = createResource( "Seq" );
                mType = createResource( "type" );
                mBag = createResource( "Bag" );
            }
            catch( ModelException e )
            {
                throw( new Exception( e.getMessage() ) );
            }
        }

        public String namespace() { return( mNamespace ); }
        public Resource object() { return( mObject ); }
        public Resource type() { return( mType ); }
        public Resource alt() { return( mAlt ); }
        public Resource subject() { return( mSubject ); }
        public Resource value() { return( mValue ); }
        public Resource predicate() { return( mPredicate ); }
        public Resource property() { return( mProperty ); }
        public Resource seq() { return( mSeq ); }
        public Resource bag() { return( mBag ); }

        private Resource createResource( String name )
            throws ModelException
        {
            return( mNodeFactory.createResource( mNamespace, name ) );
        }

        private NodeFactory mNodeFactory;

        private String mNamespace;
        private Resource mObject;
        private Resource mStatement;
        private Resource mType;
        private Resource mAlt;
        private Resource mSubject;
        private Resource mValue;
        private Resource mPredicate;
        private Resource mProperty;
        private Resource mSeq;
        private Resource mBag;
    }

    public abstract static class Import
    {
        public static class Exception extends java.lang.Exception
        {
            public Exception( String msg ) { super( msg  ); }
        }

        abstract public Map/*Resource,Object*/ from( String fileName )
            throws RDF.Import.Exception;

        abstract public Map/*Resource,Object*/ from( Reader r )
            throws RDF.Import.Exception;

        abstract public Map/*Resource,Object*/ from( Model m )
            throws RDF.Import.Exception;

        abstract public String getImportNamespace();
        abstract public String getImportPackageName();

        private static Import instantiateImportImpl( String importImplClsName,
                                                     String importNamespace,
                                                     String packageName )
           throws RDF.Import.Exception
        {
            Import imp = null;
            try
            {
                Class c = Class.forName( importImplClsName );
                Constructor ctor = c.getConstructor( new Class[] { String.class, String.class } );
                imp = (Import)ctor.newInstance( new Object[] { importNamespace, packageName } );

            }
            catch( ClassNotFoundException e )
            {
                throw( new Exception( RDF.Import.class.getName() +" implementation class '" + importImplClsName
                                       + "' not found" ) );
            }
            catch( NoSuchMethodException e )
            {
                throw( new Exception( RDF.Import.class.getName() + " implementation class '" + importImplClsName
                                     + "' does not have a feasible constructor: "
                                     + e.getMessage() ) );
            }
            catch( IllegalAccessException e )
            {
                throw( new Exception( RDF.Import.class.getName() + " implementation class '" + importImplClsName
                                     + "' does not have an accessible constructor: "
                                     + e.getMessage() ) );
            }
            catch( InstantiationException e )
            {
                throw( new Exception( RDF.Import.class.getName() + " implementation class '" + importImplClsName
                                      + "' cannot be instantiated: "
                                      + e.getMessage() ) );
            }
            catch( InvocationTargetException e )
            {
                throw( new Exception( "Exception uccured while instantiating class '" + importImplClsName
                                      + "': " + e.getMessage() ) );
            }
            catch( ClassCastException e )
            {
                throw( new Exception( "Class '" + importImplClsName
                                      + "' does not extend " + RDF.Import.class.getName() ) );
            }
            return( imp );
        }
    }

    /* Cannot be instantiated from outside */
    private RDF()
    {
    }

    public static Syntax syntax()
        throws RDF.Exception
    {
        return( syntax( null ) );
    }

    public static Syntax syntax( String namespace )
        throws RDF.Exception
    {
        try
        {
            if (namespace == null)
               namespace = Util.config().getPropertyValue( Property.DEFAULT_SYNTAX_NAMESPACE );
        }
        catch( Config.Exception e )
        {
            Util.debug().error( e );
        }

        Syntax s = (Syntax)mSyntax.get( namespace );
        if (s == null)
        {
            s = new Syntax( namespace );
            mSyntax.put( namespace, s );
        }
        return( s );
    }

    /**
     * Create a RDFFactory.
     * The default factory implementation is used.
     * @see Property#RDF_FACTORY_IMPL_CLASSNAME
     */
    public static RDFFactory factory()
        throws RDF.Exception
    {
       String clsName = null;
       try
       {
           clsName = Util.config().getPropertyValue( Property.RDF_FACTORY_IMPL_CLASSNAME );
       }
       catch( Config.Exception e )
       {
           Util.debug().error( e );
       }

       return( factory( clsName ) );
    }

    /**
     * Create a RDFFactory.
     * @param factoryImplClassName The class name of a RDFFactory implementation.
     */
    public static RDFFactory factory( String factoryImplClassName )
        throws RDF.Exception
    {
        RDFFactory f = intantiateRDFFactoryImpl( factoryImplClassName );
        return( f );
    }

    public static RDF.Import importer( String namespace,
                                       String packageName )
        throws RDF.Import.Exception
    {
        String clsName = null;
        try
        {
            clsName = Util.config().getPropertyValue( Property.IMPORT_IMPL_CLASSNAME );
        }
        catch( Config.Exception e )
        {
            Util.debug().error( e );
        }
        return( importer( clsName, namespace, packageName ) );
    }


    public static RDF.Import importer( String importerImplClassName,
                                       String namespace,
                                       String packageName )
        throws RDF.Import.Exception
    {
        return( RDF.Import.instantiateImportImpl( importerImplClassName, namespace, packageName ) );
    }


    private static RDFFactory intantiateRDFFactoryImpl( String rdfFactoryClsName )
        throws RDF.Exception
    {
        RDFFactory factory = null;

        //consistency check
        if (rdfFactoryClsName == null)
        {
            throw( new RDF.Exception( "Illegal RDF factory class name" ) );
        }

        try
        {
            factory = (RDFFactory)Objects.instantiate( rdfFactoryClsName );
        }
        catch( ClassNotFoundException e )
        {
            throw( new Exception( "RDF factory class '" + rdfFactoryClsName
                                   + "' cannot be loaded: " + e.getMessage() ) );
        }
        catch( IllegalAccessException e )
        {
            throw( new Exception( "RDF factory class '" + rdfFactoryClsName
                                 + "' does not have a zero-argument constructor: "
                                 + e.getMessage() ) );
        }
        catch( InstantiationException e )
        {
            throw( new Exception( "RDF factory class '" + rdfFactoryClsName
                                  + "' cannot be instantiated: "
                                  + e.getMessage() ) );
        }

        return( factory );
    }


    public static void parse( String uri, Reader reader, RDFParser parser, Model model )
        throws IOException,
               SAXException,
               ModelException
   {
        InputSource source = new InputSource( reader );
        source.setSystemId( uri );

        parser.parse( source, new ModelConsumer(model) );
    }

    public static void parse( String uri, Reader reader, RDFFactory factory )
        throws IOException,
               SAXException,
               ModelException
    {
        Model model = factory.createModel();
        RDFParser parser = factory.createParser();

        parse( uri, reader, parser, model );
    }

    /**
    * Creates a new unique unnamed resource.
    */
    public static Resource noname(Model m)
        throws ModelException
    {
        return( m.getNodeFactory().createUniqueResource() );
    }

    /**
     * Converts an ordinal property to an integer.
     */
    public static int getOrd(Resource r)
        throws ModelException
    {
        if(r == null)
            return( -1 );

        String uri = r.toString();
        if(!isRDF(uri))
          return( -1 );

        int pos = getNamespaceEnd(uri);
        if(pos > 0 && pos + 1 < uri.length())
        {
          try
          {
            int n = Integer.parseInt(uri.substring(pos + 1));
            if(n >= 1)
              return n;
          }
          catch (java.lang.Exception any)
          {
          }
        }
        return -1;
    }

    /**
     * Tests if the URI is qualified, i.e. has a namespace prefix.
     */
    public static boolean isQualified(String s)
    {
        int l = s.length()-1;
        do
        {
          char c = s.charAt(l);
          if(c == '#' || c == ':')
            return true;
          l--;
        }while (l >= 0);

        return false;
    }

  /**
   * Extracts the namespace prefix out of a URI.
   */
  public static String guessNamespace(String uri) {

    int l = getNamespaceEnd(uri);
    return l > 1 ? uri.substring(0, l /*-1*/) : "";
  }

  /**
   * Delivers the name out of the URI (without the namespace prefix).
   */
  public static String guessName(String uri) {

    return uri.substring(getNamespaceEnd(uri));
  }

  /**
   * Extracts the namespace prefix out of a URI.
   */
  public static String getNamespace(Resource r) throws ModelException {

    String ns = r.getNamespace();
    return ns != null ? ns : guessNamespace(r.getURI());
  }

  /**
   * Delivers the name out of the URI (without the namespace prefix).
   */
  public static String getLocalName(Resource r) throws ModelException {

    if(r.getNamespace() == null)
      return guessName(r.getURI());
    else
      return r.getLocalName();
  }

    /**
     * Tests if the URI belongs to the RDF syntax/model namespace.
     */
    public static boolean isRDF(String uri)
    {
        boolean itIs = false;
        try
        {
            itIs = (uri != null && uri.startsWith(RDF.syntax().namespace()) );
        }
        catch( Exception ignore )
        {
        }
        return( itIs );
    }

  /**
   * Tests if the resource belongs to the RDF syntax/model namespace.
   */
  public static boolean isRDF(Resource r) throws ModelException {
    return isRDF(r.getURI());
  }

  /**
   * Position of the namespace end
   */
  static int getNamespaceEnd(String uri) {

    int l = uri.length()-1;
    do {
      char c = uri.charAt(l);
      if(c == '#' || c == ':' || c == '/')
        break;
      l--;
    } while (l >= 0);
    l++;
    return l;
  }

  /**
   * Returns the first triple of the model
   */
  public static Statement get1(Model m) throws ModelException {
    if(m == null || m.isEmpty())
      return null;
    //    if(m.size() > 1)
    //      throw new RuntimeException("Model contains more than one triple");
    return (Statement)m.elements().nextElement();
  }

  public static void add(Model m, Resource subject, Resource predicate, RDFNode object) throws ModelException {

    m.add(m.getNodeFactory().createStatement(subject, predicate, object));
  }

  /**
   * returns true if old triples from r were removed
   */
  public static boolean setUniqueObject(Model r, Resource subject, Resource predicate, RDFNode object) throws ModelException {

    Model old = r.find(subject, predicate, null);
    SetOperations.subtract(r, old);
    r.add(r.getNodeFactory().createStatement(subject, predicate, object));
    return !old.isEmpty();
  }

  /**
   * returns the literal value of the node reachable from subject via predicate
   */
  public static String getObjectLiteral(Model r, Resource subject, Resource predicate) throws ModelException {

    RDFNode obj = getObject(r, subject, predicate);
    if(obj instanceof Literal)
      return obj.toString();
    else
      return null;
  }

  public static Resource getObjectResource(Model r, Resource subject, Resource predicate) throws ModelException {

    RDFNode obj = getObject(r, subject, predicate);
    if(obj instanceof Resource)
      return (Resource)obj;
    else
      return null;
  }


    public static boolean isInstanceOf(Model r, Resource i, Resource cls)
        throws ModelException
    {
        boolean itIs = false;
        try
        {
            itIs = (!r.find(i, RDF.syntax().type(), cls).isEmpty());
        }
        catch( Exception ignore )
        {
        }
        return( itIs );
  }

  public static RDFNode getObject(Model r, Resource subject, Resource predicate) throws ModelException {

    Model m = r.find(subject, predicate, null);
    if(m == null || m.size() == 0)
      return null;
    //    if(m.size() > 1)
    //      throw new RuntimeException("Model contains more than one triple");
    // FIXME: we do not check whether it is a resource or literal
    return ((Statement)m.elements().nextElement()).object();
  }

  public static Resource getSubject(Model r, Resource predicate, RDFNode object) throws ModelException {

    Model m = r.find(null, predicate, object);
    if(m == null || m.size() == 0)
      return null;
    //    if(m.size() > 1)
    //      throw new RuntimeException("Model contains more than one triple");
    return ((Statement)m.elements().nextElement()).subject();
  }

  /**
   * Prints the triples of a model to the given PrintStream.
   */
  public static void printStatements(Model m, PrintStream ps) throws ModelException {

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      ps.println (t); // "triple(\""+t.subject()+"\",\""+t.predicate()+"\",\""+t.object()+"\").");
    }
  }

  /**
   * Dumps the model in a serialized form.
   */
  public static void dumpModel(Model m, PrintStream ps, RDFSerializer s) throws ModelException, IOException, SerializationException {
    s.serialize( m, new OutputStreamWriter(ps) );
  }

  /**
   * Dumps the model in a serialized form in a string
   */
  public static String dumpModel(Model m, RDFSerializer s) throws ModelException, IOException, SerializationException {

    StringWriter w = new StringWriter();
    s.serialize( m, w );
    return w.toString();
  }

    /**
   * Dumps the model in a serialized form in a string
   */
  public static String dumpModel( Model m, Writer w, RDFSerializer s )
      throws ModelException,
             IOException,
             SerializationException
  {
    s.serialize( m, w );
    return w.toString();
  }

  /**
   * Collects the triples of a model in a vector.
   */
  public static Vector getStatementVector(Model m) throws ModelException {

    Vector v = new Vector();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      v.addElement(t);
    }
    return v;
  }

  /**
   * Collects the triples of a model into an array.
   */
  public static Statement[] getStatementArray(Model m) throws ModelException {

    Statement[] v = new Statement[m.size()];
    int i=0;
    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      v[i++] = t;
    }
    return v;
  }

  /**
   * Removes all triples which have something to do with the given namespace
   */
  public static Model removeNamespace( String ns, Model m ) throws ModelException {

    Model res = m.duplicate();
    for(Enumeration en = m.duplicate().elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      if(t.subject().toString().startsWith( ns ) ||
         t.predicate().toString().startsWith( ns ) ||
         t.object().toString().startsWith( ns )) {
        //      System.err.println("REMOVING TRIPLE: " + t);
        res.remove(t);
      }
    }
    return res;
  }

  /**
   * returns a subgraph of "m" containing
   * "r" and all nodes reachable from "r" via directed edges.
   * These edges are also included in the resulting model.
   */
  public static Model getReachable(Resource r, Model m) throws ModelException {

    Model result = m.create();
    getReachable(r, m, result);
    return result;
  }

  // FIXME: use digest instead of size
  static void getReachable(Resource r, Model m, Model result) throws ModelException {

    int oldSize = result.size();
    Model directlyReachable = m.find(r, null, null);
    SetOperations.unite(result, directlyReachable);
    if(result.size() == oldSize)
      return;
    for(Enumeration en = directlyReachable.elements(); en.hasMoreElements();) {
      Statement t = (Statement)en.nextElement();
      if(t.object() instanceof Resource)
        getReachable((Resource)t.object(), m, result);
    }
  }

  public static void parse(String fileNameOrURL, RDFParser parser, Model model)
      throws IOException,
             SAXException,
             MalformedURLException,
             ModelException
  {
    URL url = new URL (normalizeURI(fileNameOrURL));

    // maybe this model is loaded as schema...
    //    Model model = factory.registry().get(url.toString());
    //    if(model != null)
    //      return model;

    // Prepare input source
    model.setSourceURI( url.toString() );
    InputSource source = new InputSource( url.openStream() );
    source.setSystemId( url.toString() );

    parser.parse( source, new ModelConsumer(model) );
  }

  public static String normalizeURI(String uri) {

    // normalise uri
    URL url = null;
    try {
      url = new URL (uri);
    } catch (java.lang.Exception e) {
      try {
        if(uri.indexOf(':') == -1)
          url = new URL ("file", null, uri);
      } catch (java.lang.Exception e2) {}
    }
    return url != null ? url.toString() : uri;
  }

  public static Hashtable getResources(Model m) throws ModelException {

    Hashtable t = new Hashtable();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {
      Statement s = (Statement)en.nextElement();
      t.put(s.subject(), s.subject());
      if(s.object() instanceof Resource)
        t.put(s.object(), s.object());
    }
    return t;
  }

  public static Hashtable getNodes(Model m) throws ModelException {

    Hashtable t = new Hashtable();
    for(Enumeration en = m.elements(); en.hasMoreElements();) {
      Statement s = (Statement)en.nextElement();
      t.put(s.subject(), s.subject());
      t.put(s.object(), s.object());
    }
    return t;
  }

  public static void saveModel(Model m, String fileName, RDFSerializer s) throws FileNotFoundException, IOException, ModelException, SerializationException {

    //    RandomAccessFile raf = new RandomAccessFile(fileName, "rw");
    //  raf.close();
    //    FileWriter writer = new FileWriter(raf.getFD());
    FileWriter writer = new FileWriter(fileName);
    s.serialize(m, writer);
    writer.close();
  }

  /** tries to determine the file name from getSourceURI */
  public static void saveModel(Model m, RDFSerializer s) throws FileNotFoundException, IOException, ModelException, SerializationException {

    // URI to filename
    URL url = null;
    try {
      url = new URL(m.getSourceURI());
    } catch (java.lang.Exception any) {
      throw new ModelException("RDFUtil: cannot determine model file name: " + m.getSourceURI());
    }
    if("file".equals(url.getProtocol()))
      saveModel(m, url.getFile().replace('/', File.separatorChar), s);
    else
      throw new ModelException("RDFUtil: cannot save to non-file model URI: " + m.getSourceURI());
  }


  public static void collectNamespaces(Resource r, Collection target) throws ModelException {

    String ns = r.getNamespace();
    if(ns != null)
      target.add(ns);
  }

  // FIXME: what if reification is an endless loop?
  public static void collectNamespaces(Statement st, Collection target) throws ModelException {

    if(st.subject() instanceof Statement)
      collectNamespaces((Statement)st.subject(), target);
    else
      collectNamespaces(st.subject(), target);

    collectNamespaces(st.predicate(), target);

    if(st.object() instanceof Statement)
      collectNamespaces((Statement)st.object(), target);
    else if(st.object() instanceof Resource)
      collectNamespaces((Resource)st.object(), target);
  }

  public static void collectNamespaces(Model m, Collection target) throws ModelException {

    for(Enumeration en = m.elements(); en.hasMoreElements(); ) {

      Statement st = (Statement)en.nextElement();
      collectNamespaces(st, target);
    }
  }

    private static Map/*String,Syntax*/ mSyntax = new TreeMap();
}