/**
 * GNU Lesser General Public License
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation,
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Copyright:    Copyright (c) DFKI GmbH<p>
 * Company:      DFKI GmbH<p>
 * @author Andreas Lauer
 * @version 1.0
 */

package dfki.util.xml;

import java.io.PrintWriter;
import java.io.IOException;
import java.io.Writer;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringWriter;

import dfki.util.Util;
import dfki.util.config.Config;
import dfki.util.reflect.Objects;

import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;

/**
 * Class to support XML handling.
 *
 * Properties of the <code>XML</code> class are:
 * <dl>
 * <dt><b>SAX_PARSER_CLASSNAME:</b> (<code>Util.XML.SaxParserClassName</code>)</dt>
 *       <dd>Classname of  the SAX parser to be used</dd>
 * <dt><b>SAX2_PARSER_CLASSNAME:</b> (<code>Util.XML.Sax2ParserClassName</code>)</dt>
 *       <dd>Classname of  the SAX2 parser to be used</dd>
 * <dt><b>DOM_IMPL_CLASSNAME:</b> (<code>Util.XML.DOMImplClassName</code>)</dt>
 *       <dd>Name of the class implementing the abstract {@link DOM} class</dd>
 * </dl>
 * <br>
 * Look in {@link dfki.util.examples.xml} for examples.
 */
public class XML
{
    public static String MODULE_NAME = "XML";

    /**
     * Defines the propeties for the XML module.
     */
    public static class Property extends dfki.util.Util.Property
    {
        private Property( String path, String descr )
        {
            super( MODULE_NAME + Config.Property.SEPARATOR + path, descr );
        }

        public final static Property SAX_PARSER_CLASSNAME = new Property( "SaxParserClassName",
                                                                          "Classname of  the SAX parser to be used" );
        public final static Property SAX2_PARSER_CLASSNAME = new Property( "Sax2ParserClassName",
                                                                          "Classname of  the SAX2 parser to be used" );
        public final static Property DOM_IMPL_CLASSNAME = new Property( "DOMImplClassName",
                                                                          "Name of the class implementing the DOM stuff" );

        protected final static Property[] ALL_PROPERTIES = { SAX_PARSER_CLASSNAME, DOM_IMPL_CLASSNAME };
    }

    static
    {
        Setup.setDefaults();
    }

    public static class Setup
    {
        public final static String SAX_PARSER_PROPERTY = "org.xml.sax.parser";
        public final static String SAX2_PARSER_PROPERTY = "org.xml.sax.driver";

        public final static String DEFAULT_SAX_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";
        public final static String DEFAULT_SAX2_PARSER_NAME = "org.apache.xerces.parsers.SAXParser";
        public final static String DEFAULT_DOM_IMPL_CLASS_NAME = "dfki.util.xml.XercesDOMImpl";

        public static class Exception extends java.lang.Exception
        {
            public Exception( String msg ) { super( msg  ); }
        }

        static
        {
            try
            {
                //load enclosing class first
                Class.forName( XML.class.getName() );
            }
            catch( java.lang.Exception e )
            {
                Util.debug().error( e );
            }
        }

        public static Property[] allProperties()
        {
            return( Property.ALL_PROPERTIES );
        }

        public static void setDefaults()
        {
            String saxParserName = (String)System.getProperties().get( SAX_PARSER_PROPERTY );
            String sax2ParserName = (String)System.getProperties().get( SAX2_PARSER_PROPERTY );
            if (saxParserName == null)
            {
                saxParserName = DEFAULT_SAX_PARSER_NAME;
            }
            if (sax2ParserName == null)
            {
                sax2ParserName = DEFAULT_SAX2_PARSER_NAME;
            }

            Util.config().setDefaultPropertyValue( Property.SAX_PARSER_CLASSNAME, saxParserName );
            //Util.config().setDefaultPropertyValue( Property.DOM_IMPL_CLASSNAME, "dfki.util.xml.XML4JDOMImpl" );
            try
            {
                String domImplClassName = Util.config().getPropertyValue( Property.DOM_IMPL_CLASSNAME );
                if (domImplClassName == null)
                {
                    Util.config().fillInDefaultPropertyValue( Property.DOM_IMPL_CLASSNAME, DEFAULT_DOM_IMPL_CLASS_NAME );
                }
            }
            catch( Config.Exception e )
            {
                Util.debug().error( e );
            }
        }

        public static void setSAXParserClassName( String parserClassName )
        {
            Util.config().setPropertyValue( Property.SAX_PARSER_CLASSNAME, parserClassName );
            System.getProperties().put( SAX_PARSER_PROPERTY, parserClassName );
        }

        public static void setSAX2ParserClassName( String parserClassName )
        {
            Util.config().setPropertyValue( Property.SAX2_PARSER_CLASSNAME, parserClassName );
            System.getProperties().put( SAX2_PARSER_PROPERTY, parserClassName );
        }

        public static void setDOMImplClassName( String implClassName )
            throws XML.DOM.Exception
        {
             Util.config().setPropertyValue( Property.DOM_IMPL_CLASSNAME, implClassName );
             XML.DOM.instantiateDOMImpl( implClassName );
        }

        public static String info()
        {
            return "";
        }
    }

    /**
     * API to use SAX and SAX2 parsing
     */
    public static class SAX
    {
        /**
         * Defines a default error handler which simply prints
         * out the error and warning messages.
         */
        public static class DefaultErrorHandler implements ErrorHandler
        {
            public void error( SAXParseException exception )
                throws SAXException
            {
                Util.debug().error( exception );
                throw( exception );
            }

            public void fatalError( SAXParseException exception )
                throws SAXException
            {
                Util.debug().error( exception );
                throw( exception );
            }

            public void warning( SAXParseException exception )
                throws SAXException
            {
                Util.debug().message( exception.getMessage() );
                throw( exception );
            }
        }

/*
        synchronized public static ParserFactory parserFactory()
        {
        }

        synchronized public static ParserFactory parserFactory( String factoryClassName )
        {
        }

        synchronized public static XMLReaderFactory ReaderFactory()
        {
        }

        synchronized public static XMLReaderFactory ReaderFactory( String factoryClassName )
        {
        }

        synchronized public static Parser parser()
        {
        }

        synchronized public static Parser parser( String parserClassName )
        {
        }

        synchronized public static XMLReader reader()
        {
        }

        synchronized public static XMLReader reader( String readerClassName )
        {
        }
*/
    }

    /**
     * Defines a programming interface for accessing XML structures
     * represented in the document object model (DOM).
     * The interface hides implementation specific details and allows
     * for dynamic DOM implementation selection.
     *
     * Methods not depending on a particular DOM implementation
     * are made static.
     */
    public abstract static class DOM
    {
        public static class Exception extends java.lang.Exception
        {
            public Exception( String msg ) { super( msg  ); }
        }


        public abstract Document readFromURI( String fileName ) throws SAXException, IOException;
        public abstract Document readFromString( String source ) throws SAXException, IOException;
        //public abstract Document readFromFile( File f ) throws SAXException, IOException;

        /**
         * Creates a new document instance.
         */
        public abstract Document newDocument();

        /**
         * Creates an element and a subsequent text node.
         */
        public static Element createTextElement( Document ownerDocument, String elementName, String text )
        {
            Element elt = ownerDocument.createElement( elementName );
            Text textNode = ownerDocument.createTextNode( text );
            elt.appendChild( textNode );
            return( elt );
        }

        /**
         * Creates an element and a subsequent CDataSection.
         */
        public static Element createCDATAElement( Document ownerDocument, String elementName, String text )
        {
            Element elt = ownerDocument.createElement( elementName );
            CDATASection section = ownerDocument.createCDATASection( text );
            elt.appendChild( section );
            return( elt );
        }

        //Printing routines
        private static OutputFormat printFormat()
        {
            OutputFormat format = new OutputFormat( "xml", "ISO-8859-1", true );

            format.setIndenting(true);
            format.setIndent( 4 );
            format.setLineWidth( 0 );
            format.setLineSeparator( "\n" );
            format.setPreserveSpace( false );

            return( format );
        }

        /**
         * Prints the XML document to the given output stream.
         */
        public static void print( Document doc, OutputStream os )
            throws SAXException, IOException
        {
            print( doc, new OutputStreamWriter( os ) );
        }

        /**
         * Prints the XML document to the given writer.
         */
        public static void print( Document doc, Writer writer )
            throws SAXException, IOException
        {
            if (doc == null)
               return;

            OutputFormat format = printFormat();

            XMLSerializer s = new XMLSerializer( writer, format );
            s.serialize( doc );
            writer.flush();
        }

        /**
         * Prints the XML document to <code>System.out</code>.
         */
        public static void print( Document doc )
            throws SAXException, IOException
        {
            print( doc, new PrintWriter( System.out ) );
        }

        /**
         * Converts the given <code>DOM</code> structure into a string representation.
         * @return The string representation; <code>null</code> in the case of a failure.
         */
        public static String toString( Document doc )
        {
            try
            {
                StringWriter sw = new StringWriter();
                print( doc, sw );
                return( sw.toString() );
            }
            catch( java.lang.Exception e )
            {
                return( null );
            }
        }


        /**
         * Prints the XML element to the given output stream.
         */
        public static void print( Element elt, OutputStream os )
            throws SAXException, IOException
        {
            print( elt, new OutputStreamWriter( os ) );
        }

        /**
         * Prints the XML element to the given writer.
         */
        public static void print( Element elt, Writer writer )
            throws SAXException, IOException
        {
            if (elt == null)
               return;

            OutputFormat format = printFormat();

            XMLSerializer s = new XMLSerializer( writer, format );
            s.serialize( elt );
            writer.flush();
        }

        /**
         * Prints the XML element to <code>System.out</code>.
         */
        public static void print( Element elt )
            throws SAXException, IOException
        {
            print( elt, new PrintWriter( System.out ) );
        }

        /**
         * Converts the given element into a string representation.
         * @return The string representation; <code>null</code> in the case of a failure.
         */
        public static String toString( Element elt )
        {
            try
            {
                StringWriter sw = new StringWriter();
                print( elt, sw );
                return( sw.toString() );
            }
            catch( java.lang.Exception e )
            {
                Util.debug().warning( e.toString() );
                return( null );
            }
        }


        /**
         * Returns all text nodes concatenated.
         */
        public static String getText( Node node )
        {
            if (node == null)
               return( "" );

            StringBuffer buffer = new StringBuffer();
            getText(node, buffer);
            return( buffer.toString() );
        }

        /**
         * Returns all text nodes concatenated.
         */
        public static void getText( Node node, StringBuffer buffer )
        {
            if (node instanceof Text)
            {
                String text = node.getNodeValue().trim();
                if (!text.equals(""))
                {
                   if (buffer.length() > 0)
                      buffer.append(' ');
                   buffer.append(text);
                }
            }
            else if (node instanceof Element)
            {
                NodeList children = node.getChildNodes();
                for (int i = 0; i < children.getLength(); i++)
                    getText( children.item(i), buffer );
            }
        }

        /** returns null if document appears to be corrupt
         *  as the result of fatal parsing errors
         */
        protected Document checked( Document document )
        {
            if (document == null || document.getDocumentElement() == null)
              return( null );
            else
              return( document );
        }


        /**
         * Instantiate a class implementing the abstract
         * <code>DOM</code> class.
         * @see XML.DOM
         */
        private static DOM instantiateDOMImpl( String domImplClassName )
            throws XML.DOM.Exception
        {
            DOM dom = null;
            //consistency check
            try
            {
                dom = (DOM)(Class.forName( domImplClassName ).newInstance());
            }
            catch( ClassNotFoundException e )
            {
                throw( new Exception( DOM.class.getName() +" implementation class '" + domImplClassName
                                       + "' cannot be loaded " ) );
            }
            catch( IllegalAccessException e )
            {
                throw( new Exception( DOM.class.getName() + " implementation class '" + domImplClassName
                                     + "' does not have an accessible zero-argument constructor: "
                                     + e.getMessage() ) );
            }
            catch( InstantiationException e )
            {
                throw( new Exception( DOM.class.getName() + " implementation class '" + domImplClassName
                                      + "' cannot be instantiated: "
                                      + e.getMessage() ) );
            }
            catch( ClassCastException e )
            {
                throw( new Exception( "Class '" + domImplClassName
                                      + "' does not extend " + DOM.class.getName() ) );
            }

            return( dom );
        }
    }

    public abstract static class PDOM extends DOM
    {
        public Document newDocument()
        {
            return null;
        }
    }


    /**
     * Creates an instance of a <code>XML.DOM</code> implementing class.
     * The method instantiates the class set via the
     * {@link Property#DOM_IMPL_CLASSNAME} property. If no value is given
     * the default implementation is used.
     * @see Setup
     * @see DOM
     */
    public static DOM dom()
        throws XML.DOM.Exception
    {
        String clsName = null;
        try
        {
            clsName = Util.config().getPropertyValue( Property.DOM_IMPL_CLASSNAME );
        }
        catch( Config.Exception e )
        {
            Util.debug().error( e );
        }

        return( dom( clsName ) );
    }

    /**
     * Creates an instance of a <code>XML.DOM</code> implementing class.
     * @see DOM
     */
    public static DOM dom( String domImplClassName )
        throws XML.DOM.Exception
    {
        DOM d = DOM.instantiateDOMImpl( domImplClassName );
        return( d );
    }



    /**
     * Private noop constructor to prohibit instantiation from outside.
     */
    private XML()
    {
    }
}